import importlib
import os

import osw.model.entity as model
from osw.auth import CredentialManager
from osw.core import OSW
from osw.wiki_tools import SearchParam
from osw.wtsite import WtSite

# Run with:
# tox -e test -- --wiki_domain domain --wiki_username user --wiki_password pass


def test_ontology_import(wiki_domain, wiki_username, wiki_password):
    cm = CredentialManager()
    cm.add_credential(
        CredentialManager.UserPwdCredential(
            iri=wiki_domain, username=wiki_username, password=wiki_password
        )
    )
    wtsite = WtSite(WtSite.WtSiteConfig(iri=wiki_domain, cred_mngr=cm))
    # site = mwclient.Site(host=wiki_domain, scheme="http")
    # site.login(username=wiki_username, password=wiki_password)
    # wtsite = WtSite(WtSite.WtSiteLegacyConfig(site=site))
    osw_obj = OSW(site=wtsite)

    list_of_categories = [
        "Category:OSW725a3cf5458f4daea86615fcbd0029f8",  # OwlClass
        "Category:OSW6b9ef2784a934b8ab96523366e23e906",  # OwlIndividual
        "Category:Item",
        "Category:ObjectProperty",
        "Category:DataProperty",
        "Category:AnnotationProperty",
        "Category:OSW662db0a2ad0946148422245f84e82f64",  # OwlOntology
    ]
    for i, cat in enumerate(list_of_categories):
        mode = "append"
        if i == 0:
            mode = "replace"
        osw_obj.fetch_schema(OSW.FetchSchemaParam(schema_title=cat, mode=mode))

    importlib.reload(model)
    from osw.ontology import ImportConfig, OntologyImporter

    ontology_name = "example_ontology"

    ex = model.OwlOntology(
        name="Example",
        label=[model.Label(text="Example Ontology", lang="en")],
        iri="http://example.com",
        prefix="http://example.com/",
        prefix_name="example",
        see_also=["http://example.com"],
    )

    import_config = ImportConfig(
        ontology_name=ontology_name,
        ontologies=[ex],
        file=os.path.join(
            os.path.dirname(os.path.abspath(__file__)), f"{ontology_name}.ttl"
        ),
        base_class=model.OwlClass,
        base_class_title="Category:OSW725a3cf5458f4daea86615fcbd0029f8",  # OwlClass
        dump_files=True,
        dump_path=os.path.dirname(os.path.abspath(__file__)),
        dry_run=False,
    )

    # import ontologies
    importer = OntologyImporter(
        osw=osw_obj,
    )
    importer.import_ontology(import_config)

    smw_import_page = wtsite.get_page(
        WtSite.GetPageParam(titles=["MediaWiki:Smw_import_example"])
    ).pages[0]

    expected = """http://example.com/ | [http://example.com Example]
 R7k3ssL7gUxsfWuVsXWDXYF|Type:Text
 R9avr2pWFWEML712PSKDfcq|Type:Page
 REh2qNSARmKpPuwrJmr5Pu|Type:Page
 RqSw2tmyIfMbLNbk0NPkKa|Type:Text
 RBfJambxhZvFDQYeKK2zzeH|Category
 RDDfNZfAHDafrgYXW6rtT14|Category
 RDnVhTMcRkWFpWWnAprFlO0|Category
 RD2X6TQT0bKVpXehgObBb7O|Item"""

    actual = smw_import_page._content  # get_slot_content("main")
    assert actual == expected

    property_name = None
    if import_config.property_naming_policy == "label":
        property_name = "ObjectPropertyA"
    elif import_config.property_naming_policy == "prefixed_label":
        property_name = (
            f"example{import_config.property_naming_prefix_delimiter}ObjectPropertyA"
        )
    search_param = SearchParam(
        query=f"[[{property_name}::Category:OSWb1e6910f1e3d567aaed30b83ac887708]]"
    )
    full_page_titles = osw_obj.site.semantic_search(search_param)
    print(full_page_titles)
    # Not reliable, because the SMW store is not updated immediately
    # assert "Category:OSW51f195014de65ebe9f08994b21498cae" in full_page_titles

    # Not reliable due to parallel test execution using the same osw instance and
    #  page ids
    # osw_obj.delete_entity(
    #     OSW.DeleteEntityParam(
    #         entities=importer._entities, comment="[bot] delete test data"
    #     )
    # )



import os

from osw.express import OswExpress

# Create/update the password file under examples/accounts.pwd.yaml
pwd_file_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
)
# pwd_file_path = "./accounts.pwd.yaml"
osw_obj = OswExpress(
    domain="wiki-dev.open-semantic-lab.org", cred_filepath=pwd_file_path
)

# osw.fetch_schema()  # This will load the current entity schema from the OSW instance.
# You may have to re-run the script to get the updated schema extension.
# Requires 'pip install datamodel-code-generator'
DEPENDENCIES = {
    "Entity": "Category:Entity",
    "Item": "Category:Item",
    # "User": "Category:OSWd9aa0bca9b0040d8af6f5c091bf9eec7",
    # "Group": "Category:OSWb8b6278763d54b0784eea9d3b3d183a4",
    # "Location": "Category:OSW3d238d05316e45a4ac95a11d7b24e36b",
    # "Site": "Category:OSW473d7a1ed48544d1be83b258b5810948",
    # "Building": "Category:OSW4bcd4a99a73f482ea40ac4210dfab836",
    # "Floor": "Category:OSW6c4212f1a39342be963d2b9efd19c5c2",
    # "Room": "Category:OSWc5ed0ed1e33c4b31887c67af25a610c1",
}
osw_obj.install_dependencies(DEPENDENCIES, mode="append")



# This module remains as a legacy module for the time being. It will be removed in
# the future.

# import re
# from typing import Dict, List, Optional, Union
#
# from pydantic import validator
#
# from osw.model.static import OswBaseModel
# from osw.utils.strings import *



from pathlib import Path

from osw.defaults import params as default_params
from osw.defaults import paths as default_paths

# The domain, osw.express will be initialized with
default_params.wiki_domain = "wiki-dev.open-semantic-lab.org"

from osw.express import OswExpress, osw_download_file, osw_upload_file  # noqa: E402

# (Optional) Set the default credentials filepath to desired location. Otherwise,
#  it will use the default location (current working directory)
# cred_filepath_default.set_default(r"C:\Users\gold\ownCloud\Personal\accounts.pwd.yaml")

# Check setting
print(f"Credentials loaded from '{str(default_paths.cred_filepath)}'")

# The domain to connect to
domain = "wiki-dev.open-semantic-lab.org"
# domain = "demo.open-semantic-lab.org"
# Create an OswExpress object
osw_obj = OswExpress(domain=domain)

# Create a file
fp = Path("example.txt")
with open(fp, "w") as file:
    file.write("Hello, World!")
# Upload a file to an OSW instance
wiki_file = osw_upload_file(fp, domain=domain)
# Delete WikiFile from OSW instance
wiki_file.delete()
# Delete a file
fp.unlink()

# Download a file from an OSW instance and save it to a local file
local_file = osw_download_file(
    "https://wiki-dev.open-semantic-lab.org/wiki/"
    "File:OSWaa635a571dfb4aa682e43b98937f5dd3.pdf",
    overwrite=True,  # Required if file already exists
)
local_file_path = local_file.path

# Open a file with context manager directly from an OSW instance
with osw_download_file(
    "File:OSWac9224e1a280449dba71d945b1581d57.txt",
    domain=domain,
    overwrite=True,  # Required if file already exists
) as file:
    content = file.read()



from osw.utils.regex_pattern import REGEX_PATTERN_LIB

test_string = """
    {{Template:Viewer/Media
    | image_size = 300
    | mode = default
    | textdata = File:OSW420279d1be9640ad96e6685277a3f29b.png{{!}}Navigation menu;
    File:OSW0024d7c0a0d64642bf51f5facfe85d42{{!}}Search bar;
    File:OSW841b61b5996340fa81c3163bcea87482.png{{!}}Three points menu;
    File:OSWe8c6b659eab14cca927835ccd6baef15.png{{!}}Wiki preferences;
    File:OSWc0df6b35cc964307bbf3d18c78e5cb3e.png{{!}}Alerts;
    File:OSWd173625534d04fe6aab90a7bee4008e2.png{{!}}Notices;
    File:OSWc34ced55461949a59f950283e905b5fc.drawio.png{{!}}Personal menu;
    }}
    """

second_test_string = r"""
{
  "type": [
    "Category:OSW92cc6b1a2e6b4bb7bad470dfdcfdaf26"
  ],
  "author": [],
  "uuid": "d00e7453-f1f3-4a6e-94e6-3c2664fb7776",
  "label": [
    {
      "text": "Full knowledge graph",
      "lang": "en"
    }
  ],
  "description": [
    {
      "text": "Large graph displaying the full knowledge base",
      "lang": "en"
    }
  ],
  "name": "FullKnowledgeGraph",
  "image": "File:OSWc34ced55461949a59f950283e905b5fc.drawio.png"
}
"""

my_dict = {
    "type": ["Category:OSW92cc6b1a2e6b4bb7bad470dfdcfdaf26"],
    "author": [],
    "uuid": "d00e7453-f1f3-4a6e-94e6-3c2664fb7776",
    "label": [{"text": "Full knowledge graph", "lang": "en"}],
    "description": [
        {"text": "Large graph displaying the full knowledge base", "lang": "en"}
    ],
    "name": "FullKnowledgeGraph",
    "image": "File:OSWc34ced55461949a59f950283e905b5fc.drawio.png",
}

my_dict_as_str = str(my_dict)


# Test the regex pattern
# Run the following code in the Python console / this script in interactive console
my_pattern = REGEX_PATTERN_LIB["File page strings from any text"]

search_result = my_pattern.search(test_string)

findall_result = my_pattern.findall(test_string)

full_page_names = my_pattern.findall_by_group_key(test_string, "Full page name")
full_page_names_2 = my_pattern.findall_by_group_key(
    second_test_string, "Full page name"
)



"""This script provides the ability to copy a page from one OSW instance to another
OSW instance."""

from pathlib import Path

from typing_extensions import List, Optional, Union

from osw.auth import CredentialManager
from osw.core import OSW
from osw.model.static import OswBaseModel
from osw.utils import util
from osw.wtsite import SLOTS, WtPage, WtSite


class OswInstance(OswBaseModel):
    domain: str
    cred_filepath: Union[str, Path]
    credentials_manager: Optional[CredentialManager]
    osw: Optional[OSW]
    wtsite: Optional[WtSite]

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, domain: str, cred_filepath: Union[str, Path]):
        super().__init__(**{"domain": domain, "cred_filepath": cred_filepath})
        self.credentials_manager = CredentialManager(cred_filepath=cred_filepath)
        self.osw = OSW(
            site=WtSite(
                WtSite.WtSiteConfig(iri=domain, cred_mngr=self.credentials_manager)
            )
        )
        self.wtsite = self.osw.site

    def get_page_content(self, full_page_titles: List[str]) -> dict:
        get_page_res: WtSite.GetPageResult = self.wtsite.get_page(
            WtSite.GetPageParam(titles=full_page_titles)
        )

        return_dict = {}
        for page in get_page_res.pages:
            title = page.title
            slot_contents = {}
            for slot in SLOTS:
                slot_content = page.get_slot_content(slot)
                if slot_content is not None:
                    slot_contents[slot] = slot_content
            return_dict[title] = slot_contents

        return return_dict

    def set_single_page_content(
        self,
        handover_dict: dict,
        comment: str,
        overwrite: bool = False,
    ):
        full_page_title: str = handover_dict["full_page_title"]
        content_dict: dict = handover_dict["content_dict"]
        wtpage = WtPage(
            wtSite=self.wtsite,
            title=full_page_title,
        )
        if wtpage.exists:
            if overwrite is False:
                print(
                    f"Page '{full_page_title}' already exists. It will not be updated."
                )
                return {full_page_title: False}
            changed_slots = []
            for slot in SLOTS:
                remote_content = wtpage.get_slot_content(slot)
                if remote_content != content_dict.get(slot, None):
                    changed_slots.append(slot)
            if len(changed_slots) == 0:
                print(
                    f"Page '{full_page_title}' already has the same content."
                    f" It will not be updated."
                )
                return {full_page_title: False}
            else:
                print(
                    f"Page '{full_page_title}' has different content in slots "
                    f"{changed_slots}. It will be updated."
                )
        for slot in content_dict.keys():
            wtpage.create_slot(
                slot_key=slot,
                content_model=SLOTS[slot]["content_model"],
            )
            wtpage.set_slot_content(
                slot_key=slot,
                content=content_dict[slot],
            )
        wtpage.edit(
            comment=comment,
        )
        print(f"Page updated: 'https://{self.domain}/wiki/{full_page_title}'")
        return {full_page_title: True}

    def set_page_contents(
        self, content_list: List[dict], comment: str, overwrite: bool = False
    ) -> list:
        result_list = util.parallelize(
            self.set_single_page_content,
            content_list,
            comment=comment,
            overwrite=overwrite,
            flush_at_end=True,
        )
        return result_list


def copy_pages_from(
    source_domain: str,
    to_target_domains: List[str],
    page_titles: List[str],
    cred_filepath: Union[str, Path],
    comment: str = None,
    overwrite: bool = False,
):
    if comment is None:
        comment = f"[bot edit] Copied from {source_domain}"
    osw_source = OswInstance(
        domain=source_domain,
        cred_filepath=cred_filepath,
    )
    osw_targets = [
        OswInstance(
            domain=domain,
            cred_filepath=cred_filepath,
        )
        for domain in to_target_domains
    ]
    page_contents = osw_source.get_page_content(full_page_titles=page_titles)
    result = {}
    for osw_target in osw_targets:  # could also be parallelized!
        result[osw_target.domain] = osw_target.set_page_contents(
            content_list=[
                {
                    "full_page_title": full_page_title,
                    "content_dict": page_content,
                }
                for full_page_title, page_content in page_contents.items()
            ],
            comment=comment,
            overwrite=overwrite,
        )
    return result


if __name__ == "__main__":
    cred_filepath_ = Path(r"accounts.pwd.yaml")
    source = "onto-wiki.eu"
    targets = ["wiki-dev.open-semantic-lab.org"]
    titles = [
        "Item:OSW8dca6aaebe005c5faca05bac33264e4d",
        "Item:OSWaeffcee25ccb5dd8b42a434dc644d62c",
    ]
    # Implementation within this script
    use_this_script = False
    if use_this_script:
        copied_pages = copy_pages_from(
            source_domain=source,
            to_target_domains=targets,
            page_titles=titles,
            cred_filepath=cred_filepath_,
            overwrite=True,
        )
    # Implementation within wtsite
    use_wtsite = True
    if use_wtsite:
        osw_source_ = OswInstance(
            domain=source,
            cred_filepath=cred_filepath_,
        )
        osw_targets_ = [
            OswInstance(
                domain=target,
                cred_filepath=cred_filepath_,
            )
            for target in targets
        ]
        source_site = osw_source_.wtsite
        target_sites: List[WtSite] = [osw_target.wtsite for osw_target in osw_targets_]
        result_ = {}
        for target_site in target_sites:
            target = target_site.mw_site.host
            copied_pages = target_site.copy_pages(
                WtSite.CopyPagesParam(
                    source_site=source_site,
                    existing_pages=titles,
                    overwrite=True,
                    parallel=False,
                )
            )
            result_[target] = copied_pages



import uuid

from osw.auth import CredentialManager
from osw.core import OSW
from osw.wtsite import WtPage, WtSite

# run with: tox -e test -- --wiki_domain domain --wiki_username user --wiki_password pass


def _test_ontology_import(wiki_domain, wiki_username, wiki_password):
    """this test does not run with a bot account"""
    cm = CredentialManager()
    cm.add_credential(
        CredentialManager.UserPwdCredential(
            iri=wiki_domain, username=wiki_username, password=wiki_password
        )
    )
    wtsite = WtSite(WtSite.WtSiteConfig(iri=wiki_domain, cred_mngr=cm))

    osw = OSW(site=wtsite)

    p = osw.site.get_page(WtSite.GetPageParam(titles=["Main_Page"])).pages[0]
    res = p.export_xml()
    assert res.success is True

    p2 = osw.site.get_page(
        WtSite.GetPageParam(titles=["Item:" + OSW.get_osw_id(uuid.uuid4())])
    ).pages[0]
    res2 = p2.import_xml(
        WtPage.ImportConfig(
            xml=res.xml,
            summary="test import",
            source_domain="wiki-dev.open-semantic-lab.org",
        )
    )
    assert res2.success is True
    assert res2.imported_title == p2.title
    assert res2.imported_revisions > 0

    p2.delete()



import os
from uuid import UUID, uuid5

import osw.model.entity as model
from osw.core import OSW
from osw.express import OswExpress
from osw.utils.wiki import get_full_title
from osw.wtsite import WtSite

# Create/update the password file under examples/accounts.pwd.yaml
pwd_file_path = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "accounts.pwd.yaml"
)
# pwd_file_path = "./accounts.pwd.yaml"
osw_obj = OswExpress(
    domain="wiki-dev.open-semantic-lab.org", cred_filepath=pwd_file_path
)

# load dependencies
DEPENDENCIES = {
    "Tool": "Category:OSWe427aafafbac4262955b9f690a83405d",
}
osw_obj.install_dependencies(DEPENDENCIES, mode="append", policy="if-missing")

# generate a UUID namespace
EXAMPLE_UUID_NAMESPACE = UUID("bd65611d-8669-4903-8a14-af88203acc38")

sensor1 = model.Device(
    uuid=uuid5(EXAMPLE_UUID_NAMESPACE, "Sensor1"),
    name="Sensor1",
    label=[model.Label(text="Sensor 1")],
    description=[model.Description(text="A sensor")],
)

sensor2 = model.Device(
    uuid=uuid5(EXAMPLE_UUID_NAMESPACE, "Sensor2"),
    name="Sensor2",
    label=[model.Label(text="Sensor 2")],
    description=[model.Description(text="Another sensor")],
)

example_machine = model.Device(
    uuid=uuid5(EXAMPLE_UUID_NAMESPACE, "ExampleMachine"),
    name="ExampleMachine",
    label=[model.Label(text="Example Machine")],
    # note: components are modelled as statements to define
    # the relationship between the machine and the components
    components=[
        model.Component(
            # note: uuid format needs to be fixed in schema
            uuid=str(uuid5(EXAMPLE_UUID_NAMESPACE, "sensor_at_position_1")),
            component_id="sensor_at_position_1",
            label=[model.Label(text="Sensor at position 1")],
            component_instance=get_full_title(sensor1),
        ),
        model.Component(
            uuid=str(uuid5(EXAMPLE_UUID_NAMESPACE, "sensor_at_position_2")),
            component_id="sensor_at_position_2",
            label=[model.Label(text="Sensor at position 2")],
            component_instance=get_full_title(sensor2),
        ),
    ],
)

result = osw_obj.export_jsonld(
    params=OSW.ExportJsonLdParams(
        entities=[example_machine, sensor1, sensor2],
        mode=OSW.JsonLdMode.expand,
        build_rdf_graph=True,
        context_loader_config=WtSite.JsonLdContextLoaderParams(
            prefer_external_vocal=False
        ),
    )
)
graph = result.graph
# print all triples in the graph
qres = graph.query(
    """
    SELECT ?s ?p ?o
    WHERE {
    ?s ?p ?o .
    }
    """
)

print("\nAll triples in the graph:")
for row in qres:
    print(row.s, row.p, row.o)

# query all components of example_machine
qres = graph.query(
    """
    SELECT ?component ?clabel
    WHERE {
    ?s Property:HasStatement ?statement .
    ?statement Property:HasProperty Property:HasPart .
    ?statement Property:HasObject ?component .
    ?component Property:HasLabel ?clabel .
    }
    """
)
print("\nComponents of example_machine:")
for row in qres:
    print(row.clabel)

print("\n\nDefine a custom context:")

result = osw_obj.export_jsonld(
    params=OSW.ExportJsonLdParams(
        entities=[example_machine, sensor1, sensor2],
        mode=OSW.JsonLdMode.expand,
        build_rdf_graph=True,
        context={
            "ex": "http://example.org/",
            "components": "@nest",
            "component_instance": {"@id": "ex:hasPart", "@type": "@id"},
            "name": {"@id": "ex:name"},
        },
    )
)
graph = result.graph
# print all triples in the graph
qres = graph.query(
    """
    SELECT ?s ?p ?o
    WHERE {
    ?s ?p ?o .
    }
    """
)
print("\nAll triples in the graph:")
for row in qres:
    print(row.s, row.p, row.o)

# query all components of example_machine
qres = graph.query(
    """
    SELECT ?clabel
    WHERE {
    ?s <http://example.org/hasPart> ?component .
    ?component <http://example.org/name> ?clabel
    }
    """
)
print("\nComponents of example_machine:")
for row in qres:
    print(row.clabel)


print(
    "\n\ncreate a custom class with an additional property mapped in a context extension"
)


class MyCustomDeviceClass(model.Device):
    my_property: str


my_custom_device = example_machine.cast(MyCustomDeviceClass, my_property="test")

result = osw_obj.export_jsonld(
    params=OSW.ExportJsonLdParams(
        entities=[my_custom_device, sensor1, sensor2],
        mode=OSW.JsonLdMode.expand,
        build_rdf_graph=True,
        context_loader_config=WtSite.JsonLdContextLoaderParams(
            prefer_external_vocal=False
        ),
        additional_context={"my_property": "http://example.org/my_property"},
    )
)

graph = result.graph
# print all triples in the graph
qres = graph.query(
    """
    SELECT ?s ?p ?o
    WHERE {
    ?s ?p ?o .
    }
    """
)
print("\nAll triples in the graph:")
for row in qres:
    print(row.s, row.p, row.o)



from copy import deepcopy
from uuid import UUID

# Legacy imports:
from osw.model.static import get_full_title, get_namespace, get_title  # noqa: F401


def get_osw_id(uuid: UUID) -> str:
    """Generates a OSW-ID based on the given uuid by prefixing "OSW" and removing
    all '-' from the uuid-string. Duplicates OSW.get_osw_id() from src/sw/core/osw.py

    Parameters
    ----------
    uuid
        An UUID object, e.g., UUID("2ea5b605-c91f-4e5a-9559-3dff79fdd4a5")

    Returns
    -------
        OSW-ID string, e.g., OSW2ea5b605c91f4e5a95593dff79fdd4a5
    """
    return "OSW" + str(uuid).replace("-", "")


def get_uuid(osw_id) -> UUID:
    """Returns the uuid for a given OSW-ID. Duplicate of OSW.get_uuid() from src/sw/core/osw.py

    Parameters
    ----------
    osw_id
        OSW-ID string, e.g., OSW2ea5b605c91f4e5a95593dff79fdd4a5

    Returns
    -------
        uuid object, e.g., UUID("2ea5b605-c91f-4e5a-9559-3dff79fdd4a5")
    """
    return UUID(osw_id.replace("OSW", ""))


def namespace_from_full_title(full_title: str) -> str:
    """extracts the namespace from a full title (namespace:title)

    Parameters
    ----------
    full_title
        the full title to extract the namespace from

    Returns
    -------
        the namespace as a string
    """
    return full_title.replace(title_from_full_title(full_title), "").replace(":", "")


def title_from_full_title(full_title: str) -> str:
    """extracts the title from a full title (namespace:title)

    Parameters
    ----------
    full_title
        the full title to extract the title from

    Returns
    -------
        the title as a string
    """
    namespace = full_title.split(":")[0]
    return full_title.split(f"{namespace}:")[-1]


def is_empty(val):
    """checks if the given value is empty"""
    if val is None:
        return True
    elif isinstance(val, list) or isinstance(val, str) or isinstance(val, dict):
        return len(val) == 0
    return False


def remove_empty_strings(d: dict, inplace: bool = True) -> dict:
    """Iterates through the dictionary structure and removes key-value pairs
    where the value is an empty string

    Parameters
    ----------
    d:
        The dictionary to perform the operation on
    inplace:
        Whether to perform the operation in place or return a new dictionary

    Returns
    -------
    result:
        The modified dictionary
    """
    if not inplace:
        d = deepcopy(d)
    keys = list(d.keys())
    for key in keys:
        value = d[key]
        if isinstance(value, dict):
            remove_empty_strings(value)
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    remove_empty_strings(item)
        if value == "":
            del d[key]
    return d


def remove_empty(d: dict, inplace: bool = True) -> dict:
    """Iterates through the dictionary structure and removes key-value pairs
    where the value is an empty string, list, set or dictionary

    Parameters
    ----------
    d:
        The dictionary to perform the operation on
    inplace:
        Whether to perform the operation in place or return a new dictionary

    Returns
    -------
    result:
        The modified dictionary
    """
    if not inplace:
        d = deepcopy(d)
    keys = list(d.keys())
    for key in keys:
        value = d[key]
        if isinstance(value, dict):
            remove_empty(value)
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    remove_empty(item)
        if value in ["", [], {}, set()]:
            del d[key]
    return d



"""
This module is to be imported in the dynamically created and updated entity.py module.
"""

from typing import TYPE_CHECKING, Literal, Optional, Type, TypeVar, Union
from uuid import UUID
from warnings import warn

from pydantic.v1 import BaseModel, Field, constr

from osw.custom_types import NoneType
from osw.utils.strings import pascal_case

T = TypeVar("T", bound=BaseModel)

# This is dirty, but required for autocompletion:
# https://stackoverflow.com/questions/62884543/pydantic-autocompletion-in-vs-code
# Ideally, solved by custom templates in the future:
# https://github.com/koxudaxi/datamodel-code-generator/issues/860
# ToDo: Still needed?

if TYPE_CHECKING:
    from dataclasses import dataclass as _basemodel_decorator
else:
    _basemodel_decorator = lambda x: x  # noqa: E731


def custom_issubclass(obj: Union[type, T], class_name: str) -> bool:
    """
    Custom issubclass function that checks if the object is a subclass of a class
    with the given name.

    Parameters
    ----------
    obj : object
        The object to check.
    class_name : str
        The name of the class to check against.

    Returns
    -------
    bool
        True if the object is a subclass of the class with the given name,
        False otherwise.
    """

    def check_bases(cls, name):
        if hasattr(cls, "__name__") and cls.__name__ == name:
            return True
        if not hasattr(cls, "__bases__"):
            return False
        for base in cls.__bases__:
            if check_bases(base, name):
                return True
        return False

    return check_bases(obj, class_name)


def custom_isinstance(obj: Union[type, T], class_name: str) -> bool:
    """
    Custom isinstance function that checks if the object is an instance of a class with
    the given name.

    Parameters
    ----------
    obj : object
        The object to check.
    class_name : str
        The name of the class to check against.

    Returns
    -------
    bool
        True if the object is an instance of the class with the given name,
        False otherwise.
    """
    if not hasattr(obj, "__class__"):
        return False

    return custom_issubclass(obj.__class__, class_name)


@_basemodel_decorator
class OswBaseModel(BaseModel):

    class Config:
        # strict = False
        # Additional fields are allowed
        validate_assignment = True
        # Ensures that the assignment of a value to a field is validated

    def __init__(self, **data):
        if data.get("label"):
            if not isinstance(data["label"], list):
                raise ValueError(
                    "label must be a list of Label objects",
                )
            labels = []
            for label in data["label"]:
                if isinstance(label, dict):
                    labels.append(Label(**label))
                else:
                    # The list element should be a Label object
                    labels.append(label)
            data["label"] = labels
            # Ensure that the label attribute is a list of Label objects, but use
            #  custom_isinstance to avoid circular imports and ValidationError since
            #  osw.model.entity defines its own Label class
            if not all(custom_isinstance(label, "Label") for label in data["label"]):
                raise ValueError(
                    "label must be a list of Label objects",
                )
        if data.get("name") is None and "label" in data:
            data["name"] = pascal_case(data["label"][0].text)
        super().__init__(**data)

    def full_dict(self, **kwargs):  # extent BaseClass export function
        d = super().dict(**kwargs)
        for key in ("_osl_template", "_osl_footer"):
            if hasattr(self, key):
                d[key] = getattr(self, key)
                # Include selected private properties. note: private properties are not
                #  considered as discriminator
        return d

    def cast(
        self,
        cls: Union[Type[T], type],
        none_to_default: bool = False,
        remove_extra: bool = False,
        silent: bool = True,
        **kwargs,
    ) -> T:
        """Casting self into target class

        Parameters
        ----------
        cls
            target class
        kwargs
            additional attributes to be set
        none_to_default
            If True, attributes that are None will be set to their default value
        remove_extra
            If True, extra attributes that are passed to the constructor are removed
        silent
            If True, no warnings are printed
        Returns
        -------
        instance of target class
        """

        def empty_list_or_none(
            obj: Union[
                NoneType,
                list,
            ]
        ) -> bool:
            if obj is None:
                return True
            elif isinstance(obj, list):
                if len(obj) == 0:
                    return True
                elif len([item for item in obj if item is not None]) == 0:
                    return True
            return False

        combined_args = {**self.dict(), **kwargs}
        none_args = []
        if none_to_default:
            reduced = {}
            for k, v in combined_args.items():
                if empty_list_or_none(v):
                    none_args.append(k)
                else:
                    reduced[k] = v
            combined_args = reduced
        extra_args = []
        if remove_extra:
            reduced = {}
            for k, v in combined_args.items():
                if k not in cls.__fields__.keys():
                    extra_args.append(k)
                else:
                    reduced[k] = v
            combined_args = reduced
        if not silent:
            if none_to_default and none_args:
                warn(f"Removed attributes with None or empty list values: {none_args}")
            if remove_extra and extra_args:
                warn(f"Removed extra attributes: {extra_args}")
        if "type" in combined_args:
            del combined_args["type"]
        return cls(**combined_args)

    def cast_none_to_default(self, cls: Union[Type[T], type], **kwargs) -> T:
        """Casting self into target class. If the passed attribute is None or solely
        includes None values, the attribute is not passed to the instance of the
        target class, which will then fall back to the default."""

        return self.cast(cls, none_to_default=True, **kwargs)

    def get_uuid(self) -> Union[str, UUID, NoneType]:
        """Getter for the attribute 'uuid' of the entity

        Returns
        -------
            The uuid as a string or None if the uuid could not be determined
        """
        return getattr(self, "uuid", None)

    def get_osw_id(self) -> Union[str, NoneType]:
        """Determines the OSW-ID based on the entity's uuid.

        Returns
        -------
            The OSW-ID as a string or None if the OSW-ID could not be determined
        """
        return get_osw_id(self)

    def get_namespace(self) -> Union[str, NoneType]:
        """Determines the wiki namespace based on the entity's type/class

        Returns
        -------
            The namespace as a string or None if the namespace could not be determined
        """
        return get_namespace(self)

    def get_title(self) -> Union[str, NoneType]:
        """Determines the wiki page title based on the entity's data

        Returns
        -------
            The title as a string or None if the title could not be determined
        """
        return get_title(self)

    def get_iri(self) -> Union[str, NoneType]:
        """Determines the IRI / wiki full title (namespace:title) based on the entity's
        data

        Returns
        -------
            The full title as a string or None if the title could not be determined.
        """
        return get_full_title(self)


def get_osw_id(entity: Union[OswBaseModel, Type[OswBaseModel]]) -> Union[str, NoneType]:
    """Determines the OSW-ID based on the entity's data - either from the entity's
    attribute 'osw_id' or 'uuid'.

    Parameters
    ----------
    entity
        The entity to determine the OSW-ID for

    Returns
    -------
        The OSW-ID as a string or None if the OSW-ID could not be determined
    """
    osw_id = getattr(entity, "osw_id", None)
    uuid = entity.get_uuid()
    from_uuid = None if uuid is None else f"OSW{str(uuid).replace('-', '')}"
    if osw_id is None:
        return from_uuid
    if osw_id != from_uuid:
        raise ValueError(f"OSW-ID does not match UUID: {osw_id} != {from_uuid}")
    return osw_id


def get_namespace(
    entity: Union[OswBaseModel, Type[OswBaseModel]]
) -> Union[str, NoneType]:
    """Determines the wiki namespace based on the entity's type/class

    Parameters
    ----------
    entity
        The entity to determine the namespace for

    Returns
    -------
        The namespace as a string or None if the namespace could not be determined
    """
    namespace = None

    if hasattr(entity, "meta") and entity.meta and entity.meta.wiki_page:
        if entity.meta.wiki_page.namespace:
            namespace = entity.meta.wiki_page.namespace

    if namespace is None:
        if custom_issubclass(entity, "Entity"):
            namespace = "Category"
        elif custom_isinstance(entity, "Category"):
            namespace = "Category"
        elif custom_issubclass(entity, "Characteristic"):
            namespace = "Category"
        elif custom_isinstance(entity, "Item"):
            namespace = "Item"
        elif custom_isinstance(entity, "Property"):
            namespace = "Property"
        elif custom_isinstance(entity, "WikiFile"):
            namespace = "File"

    return namespace


def get_title(entity: OswBaseModel) -> Union[str, NoneType]:
    """Determines the wiki page title based on the entity's data

    Parameters
    ----------
    entity
        the entity to determine the title for

    Returns
    -------
        the title as a string or None if the title could not be determined
    """
    title = None

    if hasattr(entity, "meta") and entity.meta and entity.meta.wiki_page:
        if entity.meta.wiki_page.title:
            title = entity.meta.wiki_page.title

    if title is None:
        title = get_osw_id(entity)

    return title


def get_full_title(entity: OswBaseModel) -> Union[str, NoneType]:
    """determines the wiki full title (namespace:title) based on the entity's data

    Parameters
    ----------
    entity
        the entity to determine the full title for

    Returns
    -------
        the full title as a string or None if the title could not be determined
    """
    namespace = get_namespace(entity)
    title = get_title(entity)
    if namespace is not None and title is not None:
        return namespace + ":" + title
    elif title is not None:
        return title


class Ontology(OswBaseModel):
    iri: str
    prefix: str
    name: str
    prefix_name: str
    link: str


class Label(OswBaseModel):
    text: constr(min_length=1) = Field(..., title="Text")
    lang: Optional[Literal["en", "de"]] = Field("en", title="Lang code")



import os

from osw.auth import CredentialManager
from osw.core import model
from osw.express import OswExpress
from osw.ontology import ImportConfig, OntologyImporter

# Use credentials from file. if none are found, the user will be prompted to enter them
cm = CredentialManager(
    cred_filepath=os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
    )
)

# Create the OSW object
osw_obj = OswExpress(domain="wiki-dev.open-semantic-lab.org", cred_mngr=cm)

DEPENDENCIES = {
    "OwlClass": "Category:OSW725a3cf5458f4daea86615fcbd0029f8",
    "OwlIndividual": "Category:OSW6b9ef2784a934b8ab96523366e23e906",
    "Item": "Category:Item",
    "ObjectProperty": "Category:ObjectProperty",
    "DataProperty": "Category:DataProperty",
    "AnnotationProperty": "Category:AnnotationProperty",
}
osw_obj.install_dependencies(DEPENDENCIES, mode="append")

ontology_name = "example_ontology"

ex = model.OwlOntology(
    name="Example",
    label=[model.Label(text="Example", lang="en")],
    iri="http://example.com",
    prefix="http://example.com/",
    prefix_name="example",
    see_also=["http://example.com"],
)

import_config = ImportConfig(
    ontology_name=ontology_name,
    ontologies=[ex],
    file=os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "data", f"{ontology_name}.ttl"
    ),
    base_class=model.OwlClass,
    base_class_title="Category:OSW725a3cf5458f4daea86615fcbd0029f8",  # OwlClass
    dump_files=True,
    dump_path=os.path.dirname(os.path.abspath(__file__)),
    dry_run=False,
)

# Import ontologies
importer = OntologyImporter(
    osw=osw_obj,
)
importer.import_ontology(import_config)



import mwclient

import osw.model.entity as model
from osw.auth import CredentialManager
from osw.core import OSW
from osw.wtsite import WtSite

# run with: tox -e test -- --wiki_domain domain --wiki_username user
# --wiki_password pass --db_username user --db_password pass


def _test_connection_string(
    wiki_domain, wiki_username, wiki_password, db_username, db_password
):
    # make sure to import controllers after updating the model (ignore linter warning)
    import osw.controller as controller  # noqa: E402

    cstr = controller.DatabaseController.ConnectionString(
        dialect="postgresql",
        driver="pg8000",
        username="dbuser",
        password="kx@jj5/g",
        host="pghost10",
        port="3000",
        database="appdb",
    )
    assert str(cstr) == "postgresql+pg8000://dbuser:kx%40jj5%2Fg@pghost10:3000/appdb"

    cstr = controller.DatabaseController.ConnectionString(
        dialect="postgresql",
        username="dbuser",
        password="kx@jj5/g",
        host="pghost10",
        port="3000",
        database="appdb",
    )
    assert str(cstr) == "postgresql://dbuser:kx%40jj5%2Fg@pghost10:3000/appdb"


def test_connect_and_query(
    wiki_domain, wiki_username, wiki_password, db_username, db_password
):
    site = mwclient.Site(host=wiki_domain)
    site.login(username=wiki_username, password=wiki_password)
    wtsite = WtSite(WtSite.WtSiteLegacyConfig(site=site))
    osw = OSW(site=wtsite)

    osw.fetch_schema(
        OSW.FetchSchemaParam(
            schema_title=[
                "Category:OSW02590972aeba46d7864ed492c0c11384",  # Host
                "Category:OSWacdb001c926c46b998af3e645d36b13f",  # DatabaseServer
                "Category:OSW51ad0d1716254c77a2b7a03217f23b43",  # Database
                "Category:OSW33c7c3a4076a4a58b20d818c162e84e6",  # DatabaseType
            ],
            mode="replace",
        )
    )

    # make sure to import controllers after updating the model (ignore linter warning)
    # the following does not work reliably if module was imported before model was loaded
    # import osw.controller as controller
    from osw.controller.database import DatabaseController  # noqa: E402

    # load database definition
    db = osw.load_entity("Item:OSWb8cc7705e17c47b19331fdb045bfbca8")  # postgres
    db = db.cast(model.Database)

    # cast into controller and execute function
    db = db.cast(DatabaseController)
    db.connect(
        DatabaseController.ConnectionConfig(
            cm=CredentialManager.UserPwdCredential(
                iri="", username=db_username, password=db_password
            ),
            osw=osw,
        )
    )
    db.execute("SELECT * FROM public.sensors ORDER BY sensor_id ASC")



from __future__ import annotations

import re
from pathlib import Path
from typing import List, Union

from pydantic.v1 import PrivateAttr, validator

from osw.model.static import OswBaseModel

PACKAGE_ROOT_PATH = Path(__file__).parents[1]
BASE_PATH = Path.cwd()
OSW_FILES_DIR_DEFAULT = BASE_PATH / "osw_files"
DOWNLOAD_DIR_DEFAULT = OSW_FILES_DIR_DEFAULT / "downloads"
CRED_FILENAME_DEFAULT = "accounts.pwd.yaml"
CRED_FILEPATH_DEFAULT = OSW_FILES_DIR_DEFAULT / CRED_FILENAME_DEFAULT
WIKI_DOMAIN_DEFAULT = "wiki.open-semantic-lab.org"


class FilePathDefault(OswBaseModel):
    """A class to store the default file path. This is a helper class to make the
    default file path, defined within this module, accessible from a calling script."""

    _default: Union[Path] = PrivateAttr(BASE_PATH)

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, path: Union[str, Path] = BASE_PATH):
        data = {"_default": Path(path) if isinstance(path, str) else path}
        super().__init__(**data)

    def __str__(self):
        return str(self.path)

    def __repr__(self):
        return f"FilePathDefault(path={self.path})"

    def __eq__(self, other):
        other_path = getattr(other, "path", None)
        if other_path is None:
            return False

    @property
    def path(self):
        return self._default

    def set(self, new: Union[str, Path]):
        self._default = Path(new) if isinstance(new, str) else new

    def get(self):
        return self._default


class Paths(OswBaseModel):
    """A class to store the default paths. This is a helper class to make the default
    paths, defined within this module, accessible from a calling script."""

    base: Path = BASE_PATH
    """If you want to have the sub folders created in an directory that is not the
    current working directory of the calling script, use Path.base = new_path."""
    osw_files_dir: Path = OSW_FILES_DIR_DEFAULT
    """If you want to specify the default OSW files directory, use
    Path.osw_files_dir = new_path."""
    cred_filepath: Path = CRED_FILEPATH_DEFAULT
    """If you want to specify the saving location of the credentials file, use
    Path.cred_filepath = new_path."""
    download_dir: Path = DOWNLOAD_DIR_DEFAULT
    """If you want to specify the default download directory, use
    Path.download_dir = new_path."""
    _changed: List[str] = PrivateAttr(default_factory=list)
    """A flag to indicate if any of the paths have been changed."""

    def __init__(self, **data):
        super().__init__(**data)
        self._changed = [key for key in data.keys() if key != "_changed"]

    def __setattr__(self, name, value):
        old_value = getattr(self, name)
        super().__setattr__(name, value)
        self.on_attribute_set(name, value, old_value)
        if name not in self._changed:
            self._changed.append(name)

    def has_changed(self, name):
        return name in self._changed

    @property
    def changed(self):
        return self._changed

    def on_attribute_set(self, attr_name, new_value, old_value):
        """
        This method is called every time an attribute is set. It is used to update the
        paths that are dependent on the [base, osw_files_dir] path when the
        [base, osw_files_dir] path is changed.

        Parameters
        ----------
        attr_name:
            The name of the attribute that was set
        new_value
            The new value the attribute was set to
        old_value
            The value the attribute had before it was set
        """

        def update_attr(set_attr, to_update, old_val, new_val):
            for attr_name in to_update:
                if old_val in getattr(self, attr_name).parents:
                    old_rel_path = getattr(self, attr_name).relative_to(old_val)
                    new_rel_path = new_val / old_rel_path
                    setattr(self, attr_name, new_rel_path)
                    print(
                        f"Following the setting of {self.__name__}.{set_attr}, "
                        f"{self.__name__}.{attr_name} was updated to {new_rel_path}."
                    )

        if attr_name == "base":
            update_attr(
                "base",
                ["osw_files_dir", "cred_filepath", "download_dir"],
                old_value,
                new_value,
            )
        elif attr_name == "osw_files_dir":
            update_attr(
                "osw_files_dir", ["cred_filepath", "download_dir"], old_value, new_value
            )


# Create an instance of the Paths class to make the default paths accessible from a
# calling script.
paths = Paths()
"""To overwrite a default path, execute, e.g.,: paths.base = new_path"""


class Params(OswBaseModel):
    """A class to store the default parameters. This is a helper class to make the
    default parameters, defined within this module, accessible from a calling script."""

    wiki_domain: str = WIKI_DOMAIN_DEFAULT
    """The default domain of the OSW instance to interact with."""
    _changed: List[str] = PrivateAttr(default_factory=list)
    """A flag to indicate if any of the parameters have been changed."""

    def __init__(self, **data):
        super().__init__(**data)
        self._changed = [key for key in data.keys() if key != "_changed"]

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if name != "_changed" and name not in self._changed:
            self._changed.append(name)

    @validator("wiki_domain")
    def validate_wiki_domain(cls, v):
        pattern = r"^(?!-)[A-Za-z0-9.-]{1,63}(?<!-)\.[A-Za-z]{2,}$"
        assert re.match(pattern, v), "The wiki domain is not valid."
        return v

    def has_changed(self, name):
        return name in self._changed

    @property
    def changed(self):
        return self._changed


params = Params()



import sys
import uuid as uuid_module
from pathlib import Path

import mwclient

import osw.wiki_tools as wt
from osw.data.import_utility import uuid_to_full_page_title
from osw.wtsite import WtPage, WtSite

# run with: tox -e test -- --wiki_domain domain --wiki_username user --wiki_password pass
# run locally with: tox -e test -- --wiki_domain None --wiki_username None --wiki_password None
# This will trigger the fallback option (and not the skip condition)


# Make non-package scripts available for import
cwd = Path(__file__).parent.absolute()
migration_import_dir = cwd.parents[1] / "scripts" / "migration"
sys.path.append(str(migration_import_dir))
# integration_import_dir = cwd.parents[1] / "tests" / "integration"
# sys.path.append(str(integration_import_dir))


def test_dependencies(wiki_domain: str, wiki_username: str, wiki_password: str):
    """This test makes sure that all dependencies are available. First the dependencies
    are fetched from the wiki if necessary, then the dependencies are checked."""
    import file_page_migration_dependencies as fpmd

    fpmd.main(wiki_domain, wiki_username, wiki_password)
    assert fpmd.check_dependencies()


def test_file_page_migration(wiki_domain: str, wiki_username: str, wiki_password: str):
    from file_page_migration import (
        PWD_FILE_PATH,
        WikiFile,
        move_file_pages_and_redirect_references,
    )

    # For local testing without tox
    if wiki_domain is None or wiki_domain == "None":
        wtsite = WtSite.from_domain(
            domain="wiki-dev.open-semantic-lab.org",
            password_file=str(PWD_FILE_PATH),
        )
    # For testing with tox
    else:
        site = mwclient.Site(host=wiki_domain)
        site.login(username=wiki_username, password=wiki_password)
        wtsite = WtSite(WtSite.WtSiteLegacyConfig(site=site))

    # Path to the test file:
    bare_file_name = "test"
    file_name = (
        bare_file_name + ".svg"
    )  # OSW.get_osw_id(uuid_module.uuid4()) # + ".svg"
    # Using page
    using_page_fpt = uuid_to_full_page_title(uuid_module.uuid4())
    # File test page
    file_page_name = using_page_fpt.replace(":", "-") + "-" + file_name
    file_fpt = "File:" + file_page_name
    path_to_test_file = cwd / "test.svg"
    # Make sure the state is clean - delete the file if it exists, otherwise create it
    file_page = wtsite.get_page(WtSite.GetPageParam(titles=[file_page_name])).pages[0]
    if file_page.exists:
        file_page.delete()
    with open(path_to_test_file, "r") as file:
        _ = wtsite.mw_site.upload(
            file=file,
            filename=file_page_name,
            comment="File for testing the file_page_migration.py script. "
            "Can be deleted.",
            description="Test file",
            ignore=True,
        )
    # Make sure the state is clean - delete the using page if it exists, otherwise
    #  create it
    using_page = WtPage(wtSite=wtsite, title=using_page_fpt)
    if using_page.exists:
        using_page.delete()
    using_page.set_slot_content(
        slot_key="main",
        content="{{Template:Editor/SvgEdit|file_name="
        + bare_file_name
        + "| page_name="
        + using_page_fpt
        + "}}",
    )
    using_page.edit(comment="Test")
    print("Using page created: " + using_page_fpt)
    ###############################################
    # Test the migration (serial and parallel) # todo: test parallel
    result = move_file_pages_and_redirect_references(
        file_pages=[file_fpt], wtsite_obj=wtsite, parallel=False
    )
    ###############################################
    new_full_page_name = result[0]
    new_file_page = wtsite.get_page(
        WtSite.GetPageParam(titles=[new_full_page_name])
    ).pages[0]
    # Assert that the result is as expected
    exception = None
    try:
        # Make sure the changes are propagated / Caches are invalidated
        # new_file_page.purge()
        # Get the new file page after purging
        # new_file_page = wtsite.get_page(
        #     WtSite.GetPageParam(titles=[new_full_page_name])
        # ).pages[0]
        print(f"Fetching jsondata of {new_full_page_name}")
        assert new_file_page.exists
        assert new_file_page.is_file_page
        jsondata = new_file_page.get_slot_content("jsondata")
        if jsondata is None:
            raise ValueError(
                "jsondata is None - meta data was not written to the file " "page!"
            )
        assert jsondata["type"] == WikiFile.__fields__["type"].get_default()
        modified_using_page = wtsite.get_page(
            WtSite.GetPageParam(titles=[using_page_fpt])
        ).pages[0]
        main_slot_content = modified_using_page.get_slot_content("main")
        parsed_content = wt.create_flat_content_structure_from_wikitext(
            main_slot_content
        )
        content_dict = parsed_content[0]
        assert bare_file_name in content_dict["Template:Editor/SvgEdit"]["file_name"]
        assert "uuid" in list(content_dict.values())[0].keys()
    except AssertionError as e:
        # Assertion Error for now
        exception = e
    finally:
        # Make sure the state is clean - delete the created pages
        # Delete the using page
        using_page.delete(comment="Test finished")
        # Delete the file page
        new_file_page.delete(comment="Test finished")
        # Now raise the exception if there was one
        if exception is not None:
            raise exception


# Dummy call to avoid warning
# helper.foobar()
if __name__ == "__main__":
    test_file_page_migration()



import os
import uuid

import osw.model.entity as model
from osw.auth import CredentialManager
from osw.core import OSW
from osw.express import OswExpress
from osw.utils.wiki import get_full_title, get_uuid

# Credential manager(can use a file or hardcode the credentials, otherwise the user
#  will be prompted to enter them)
cm = CredentialManager(
    cred_filepath=os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
    )
    # CredentialManager.UserPwdCredential(
    #     iri=wiki_domain, username=wiki_username, password=wiki_password
    # )
)
# Change the domain to your osw instance
osw_obj = OswExpress(domain="wiki-dev.open-semantic-lab.org", cred_mngr=cm)

# Load the required schemas / data classes
force_schema_reload = False
if force_schema_reload or not hasattr(
    model, "Corporation"
):  # only load if not already loaded. note: does not detect updated schemas yet
    DEPENDENCIES = {
        "MetaDeviceCategory": "Category:OSWd845b96813a344458f140e48c4d063fd",
        "Device(SubclassWithMetaModel)": "Category:OSW5bf1542d9cf847db83cbc73d579ba9d6",
        "Corporation": "Category:OSW5f4a3751d23e482d80fb0b72dcd6bc31",  #
    }
    osw_obj.install_dependencies(DEPENDENCIES, mode="replace")

# To reproduce your import you should define a UUIDv5 namespace
# Example: id of the source dataset / file
uuid_namespace = get_uuid("OSWc0dda5f5f066452dbab68cc8d5dcb022")

# Create a new manufacturer
new_manufacturer = model.Corporation(
    uuid=uuid.uuid5(
        uuid_namespace, "MyNewManufacturer"
    ),  # use a stable id from the source dataset / file
    name="MyNewManufacturer",
    label=[model.Label(text="My New Manufacturer", lang="en")],
    description=[model.Description(text="This is a test manufacturer", lang="en")],
)

# Create a new device type with the new manufacturer as manufacturer
new_category = model.MetaDeviceCategory(  # here used as a device type
    uuid=uuid.uuid5(
        uuid_namespace, "MyNewDeviceCategory"
    ),  # use a stable id from the source dataset / file
    subclass_of=[
        "Category:OSW5bf1542d9cf847db83cbc73d579ba9d6"
    ],  # Device (SubclassWithMetaModel)
    name="MyNewDeviceCategory",
    label=[model.Label(text="My New Category", lang="en")],
    description=[model.Description(text="This is a test device category", lang="en")],
    manufacturer=get_full_title(new_manufacturer),
)

# Store the manufacturer as an item
osw_obj.store_entity(OSW.StoreEntityParam(entities=[new_manufacturer]))
# Store the device type as a category. note: we have to request explicitly to create
#  a new category/class from this meta class instance
osw_obj.store_entity(
    OSW.StoreEntityParam(
        entities=[new_category],
        namespace="Category",
        # meta_category_title= (
        # "Category:OSWd845b96813a344458f140e48c4d063fd")
        # usage of MetaDeviceCategory not yet supported
    )
)

# Reload the schema (currently there is no direct way)
if force_schema_reload or not hasattr(
    model, "MyNewDeviceCategory"
):  # Only load if not already loaded. note: does not detect updated schemas yet
    osw_obj.fetch_schema(
        OSW.FetchSchemaParam(
            schema_title=[
                get_full_title(new_category),  # MyNewDeviceCategory
            ],
            mode="append",
        )
    )

# Create a new device of the new device type
new_instance = model.MyNewDeviceCategory(
    uuid=uuid.uuid5(
        uuid_namespace, "MyNewDevice"
    ),  # use a stable id from the source dataset / file
    name="MyNewDevice",
    label=[model.Label(text="My New Device", lang="en")],
    description=[model.Description(text="This is a test device", lang="en")],
    serial_number="1234567890",
)

# Store the device instance as an item
osw_obj.store_entity(OSW.StoreEntityParam(entities=[new_instance]))



from osw.express import osw_download_file

# Download a file from an OSW instance and save it to a local file
local_file = osw_download_file(
    "https://wiki-dev.open-semantic-lab.org/wiki/"
    "File:OSWaa635a571dfb4aa682e43b98937f5dd3.pdf"
    # , use_cached=True
    # , overwrite=True
)
local_file_path = local_file.path



# Functions
def pascal_case(st: str) -> str:
    """converts a string to PascalCase

    Parameters
    ----------
    st
        the string to convert to PascalCase

    Returns
    -------
        The string in PascalCase
    """
    if not st.isalnum():
        st = "".join(x for x in st.title() if x.isalnum())
    return st[0].upper() + st[1:]



import time

import pytest

import osw.model.entity as model
from osw.auth import CredentialManager
from osw.utils.wiki import get_full_title
from osw.wtsite import WtSite


def test_fetch_and_load(wiki_domain, wiki_username, wiki_password, mocker):
    # create a credential file on the default path for osw express
    cm = CredentialManager()  # cred_filepath = Path.cwd() / "accounts.pwd.yaml")
    cm.add_credential(
        CredentialManager.UserPwdCredential(
            iri=wiki_domain, username=wiki_username, password=wiki_password
        )
    )
    # cm.save_credentials_to_file()
    # wtsite = WtSite(WtSite.WtSiteConfig(iri=wiki_domain, cred_mngr=cm))
    # osw_obj = OSW(site=wtsite)

    # Here the initial connection to the wiki is mocked (passing domain, username and
    # password)
    if mocker is not None:
        mocked_input = mocker.patch("builtins.input")
        mocked_getpass = mocker.patch("getpass.getpass")
        mocked_input.side_effect = [wiki_domain, wiki_username]
        mocked_getpass.return_value = wiki_password
    # This import will trigger the install_dependencies method call on the first run
    import osw.express

    osw_express = osw.express.OswExpress(domain=wiki_domain, cred_mngr=cm)

    # Measure the time taken to store and load an entity
    start_time = time.time()

    # Load Tutorial Schema on demand
    # if not hasattr(model, "Tutorial"):
    DEPENDENCIES = {"Tutorial": "Category:OSW494f660e6a714a1a9681c517bbb975da"}
    osw_express.install_dependencies(DEPENDENCIES, policy="force")

    end_time = time.time()
    print(f"Time taken to load Tutorial Schema: {end_time - start_time}")
    assert hasattr(model, "Tutorial")
    assert end_time - start_time < 32  # typically takes 8 seconds

    start_time = time.time()
    osw_express.install_dependencies(DEPENDENCIES, policy="force")
    end_time = time.time()
    print(
        f"Time taken to reload Tutorial Schema with policy 'force': {end_time - start_time}"
    )
    assert end_time - start_time < 4  # typically takes 1 seconds using memory cache

    start_time = time.time()
    osw_express.install_dependencies(DEPENDENCIES, policy="if-missing")
    end_time = time.time()
    print(
        "Time taken to reload Tutorial Schema with policy ",
        f"'if-missing': {end_time - start_time}",
    )
    assert end_time - start_time < 0.5  # typically takes 0 seconds

    # query any 50 Category pages
    pages = osw_express.query_instances("Category:Category")
    start_time = time.time()
    result = osw_express.site.get_page(WtSite.GetPageParam(titles=pages[:50]))
    end_time = time.time()
    print(f"Time taken to fetch 50 Category pages: {end_time - start_time}")
    assert (
        end_time - start_time < 15
    )  # typically takes 3.5 (dask) or 2.7 (asyncio) seconds

    assert len(result.pages) == 50

    # we should get a warning for the non-existent page
    # see https://docs.pytest.org/en/stable/how-to/capture-warnings.html
    pages[0] = "IDONOTEXIST"
    with pytest.warns(RuntimeWarning) as record:
        result = osw_express.site.get_page(
            WtSite.GetPageParam(titles=pages[:50], raise_exception=False)
        )
        # assert that at least one warning message contains
        # the missing page title to inform the user
        assert any("IDONOTEXIST" in str(rec.message) for rec in record)
        assert len(result.pages) == 50  # assert that the result is still returned

    # assert that ValueError is raised when the page does not exist
    pages[0] = "IDONOTEXIST"
    with pytest.raises(ValueError):
        result = osw_express.site.get_page(
            WtSite.GetPageParam(titles=pages[:50], raise_exception=True)
        )


def _test_store(wiki_domain, wiki_username, wiki_password, mocker):
    """uploads entities and then fetches them back to compare the UUIDs
    expensive test, not enabled by default
    """

    # create a credential file on the default path for osw express
    cm = CredentialManager()  # cred_filepath = Path.cwd() / "accounts.pwd.yaml")
    cm.add_credential(
        CredentialManager.UserPwdCredential(
            iri=wiki_domain, username=wiki_username, password=wiki_password
        )
    )
    # cm.save_credentials_to_file()
    # wtsite = WtSite(WtSite.WtSiteConfig(iri=wiki_domain, cred_mngr=cm))
    # osw_obj = OSW(site=wtsite)

    # Here the initial connection to the wiki is mocked (passing domain, username and
    # password)
    if mocker is not None:
        mocked_input = mocker.patch("builtins.input")
        mocked_getpass = mocker.patch("getpass.getpass")
        mocked_input.side_effect = [wiki_domain, wiki_username]
        mocked_getpass.return_value = wiki_password
    # This import will trigger the install_dependencies method call on the first run
    import osw.express

    osw_express = osw.express.OswExpress(domain=wiki_domain, cred_mngr=cm)

    # create 50 Item entities
    item_count = 50
    items = []
    for i in range(item_count):
        item = model.Item(label=[model.Label(text=f"Performance Test Item {i}")])
        items.append(item)

    # Measure the time taken to store and load an entity
    start_time = time.time()
    # store the entities
    osw_express.store_entity(
        osw.express.OswExpress.StoreEntityParam(entities=items, parallel=True)
    )

    end_time = time.time()
    print(f"Time taken to store {item_count} Item entities: {end_time - start_time}")
    assert end_time - start_time < 60  # typically takes 30 seconds

    # load the entities
    # generate the full titles from the item objects
    item_titles = [get_full_title(item) for item in items]
    start_time = time.time()
    entities = osw_express.load_entity(
        osw.express.OswExpress.LoadEntityParam(titles=item_titles, parallel=True)
    ).entities
    end_time = time.time()
    print(f"Time taken to load {item_count} Item entities: {end_time - start_time}")
    assert end_time - start_time < 40  # typically takes 20 seconds

    # clean up
    start_time = time.time()
    osw_express.delete_entity(
        osw.express.OswExpress.DeleteEntityParam(entities=items, parallel=True)
    )
    end_time = time.time()
    print(f"Time taken to delete {item_count} Item entities: {end_time - start_time}")
    assert end_time - start_time < 20  # typically takes 10 seconds

    assert len(entities) == item_count
    # compare the loaded entities with the stored entities
    # print UUIDs in items and entities
    for i in range(item_count):
        print(f"Item {i} UUID: {items[i].uuid} Entity {i} UUID: {entities[i].uuid}")

    # compare the loaded entities with the stored entities per UUID
    # order is not guaranteed
    stored_uuids = [item.uuid for item in items]
    loaded_uuids = [entity.uuid for entity in entities]
    assert set(stored_uuids) == set(loaded_uuids)



from io import BytesIO, StringIO
from typing import IO, Any, Dict, Optional

import boto3

from osw.auth import CredentialManager
from osw.controller.file.remote import RemoteFileController
from osw.core import model


class S3FileController(model.S3File, RemoteFileController):
    protocol: Optional[str]
    domain: Optional[str]
    bucket: Optional[str]
    key: Optional[str]

    cm: CredentialManager
    s3_client: Optional[Any]
    s3_resource: Optional[Any]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._parse_url()
        creds: CredentialManager.UserPwdCredential = self.cm.get_credential(
            CredentialManager.CredentialConfig(iri=self.domain)
        )

        self.s3_resource = boto3.resource(
            "s3",
            aws_access_key_id=creds.username,
            aws_secret_access_key=creds.password,
            aws_session_token=None,  # replace this with token if necessary
            endpoint_url=self.protocol + "//" + self.domain,
            config=boto3.session.Config(signature_version="s3v4"),
            # verify=False
        )
        self.s3_client = self.s3_resource.meta.client

    def get(self) -> IO:
        response = self.s3_resource.Object(bucket_name=self.bucket, key=self.key).get()
        return response["Body"]

    # def get_to(self, other: "FileController"):
    #    response = self.s3_resource.Object(bucket_name=self.bucket, key=self.key).get()
    #    with response['Body'] as file:
    #        other.put(file)

    def put(self, file: IO, **kwargs: Dict[str, Any]):
        if isinstance(file, StringIO):
            file = BytesIO(file.getvalue().encode())
            # file.seek(0)
        self.s3_client.upload_fileobj(file, self.bucket, self.key)

    # def put_from(self, other: FileController):
    #    pass

    def delete(self):
        self.s3_client.delete_object(Bucket=self.bucket, Key=self.key)

    def _parse_url(self):
        self.protocol = self.url.split("//")[0]
        self.domain = self.url.split("//")[1].split("/")[0]
        self.bucket = self.url.split("//")[1].split("/")[1]
        self.key = self.url.replace(
            self.protocol + "//" + self.domain + "/" + self.bucket + "/", ""
        )



import shutil
from io import StringIO
from typing import IO, Any, Dict, List, Optional

from osw.controller.file.base import FileController
from osw.core import model


class InMemoryController(FileController, model.LocalFile):
    """File controller for local files"""

    label: Optional[List[model.Label]] = [model.Label(text="Unnamed stream")]
    """the label of the stream, e.g., the name of the file the stream
    originates from. Defaults to 'Unnamed stream'."""
    stream: IO
    """the stream to the file"""

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **kwargs):
        self.stream = StringIO()
        super().__init__(**kwargs)

    def get(self) -> IO:
        return self.stream

    def put(self, file: IO, **kwargs: Dict[str, Any]):
        shutil.copyfileobj(file, self.stream)



"""
This test can't be run directly (anymore) since calling fixtures directly has been
deprecated in pytest 4.0.0. Instead, the test can be run with the following command,
which requires pytest-mock to be installed:
tox -e test -- path/to/test_express.py --wiki_domain domain --wiki_username user
--wiki_password pass


Tests to be written for express.py:
* test_download_file
    * url specified
    * only title specified -> domain requested
    * target_dir (not) specified
    * target_fn (not) specified
    * target_fp (not) specified
    * osw_express (not) specified -> domain requested
    * domain (not) specified -> cred_filepath requested
    * cred_filepath (not) specified -> use fallback
    * credentials file (not) found -> check credentials_manager
    * cred_mngr (not) specified -> request credentials
    * test return value
* test_upload_file
"""

from contextlib import contextmanager
from pathlib import Path

import yaml

# requires pytest_mock fixture --> pip install pytest-mock


def create_credentials_file(
    cred_filepath: Path, wiki_domain: str, wiki_username: str, wiki_password: str
):
    with open(cred_filepath, "w") as f:
        yaml.dump(
            {wiki_domain: {"username": wiki_username, "password": wiki_password}}, f
        )


def create_dummy_file(file_path: Path):
    with open(file_path, "w") as file:
        file.write("Hello, World!")


@contextmanager
def preserve_entity_py_state():
    path = Path(__file__).parents[2] / "src" / "osw" / "model" / "entity.py"
    with open(path, "r") as f:
        original_entity = f.read()
    try:
        yield None
    finally:
        with open(path, "w") as f:
            f.write(original_entity)


def osw_express_and_credentials(osw_express, wiki_domain, wiki_username, wiki_password):
    assert osw_express.domain == wiki_domain
    # An error would occur if the connection attempt to the wiki would have failed
    assert osw_express.site is not None
    assert osw_express.cred_filepath is not None
    assert osw_express.cred_filepath.exists()
    assert osw_express.cred_filepath.is_file()
    credentials = yaml.safe_load(osw_express.cred_filepath.open("r"))
    assert wiki_domain in credentials
    assert credentials[wiki_domain]["username"] == wiki_username
    assert credentials[wiki_domain]["password"] == wiki_password
    return True


def test_init_with_domain(wiki_domain, wiki_username, wiki_password, mocker):
    """Test OswExpress initialization with defined domain, but no cred_filepath nor
    cred_mngr. As this is the first test to load the osw.express module,
    the installation of the dependencies should be triggered here."""
    # Here the initial connection to the wiki is mocked (passing domain, username and
    # password)
    mocked_input = mocker.patch("builtins.input")
    mocked_getpass = mocker.patch("getpass.getpass")
    mocked_input.side_effect = [wiki_domain, wiki_username]
    mocked_getpass.return_value = wiki_password

    # Before making changes to osw.model.entity save the original state
    with preserve_entity_py_state():
        # This import will trigger the install_dependencies method call on the first run
        import osw.express

        # A second connection is then opened with domain already set, so mocking is
        # required for the username and password only
        mocked_input.side_effect = [wiki_username]
        mocked_getpass.return_value = wiki_password
        osw_express = osw.express.OswExpress(domain=wiki_domain)
        osw_express_and_credentials(
            osw_express, wiki_domain, wiki_username, wiki_password
        )
        assert osw_express.cred_filepath == Path(
            osw.express.default_paths.cred_filepath
        )
        osw_express.shut_down()
        osw_express.cred_filepath.unlink()


def test_init_with_defaults_set(wiki_domain, wiki_username, wiki_password):
    from osw.defaults import params as default_params
    from osw.defaults import paths as default_paths

    cred_filepath = Path.cwd() / "accounts.pwd.yaml"
    create_credentials_file(cred_filepath, wiki_domain, wiki_username, wiki_password)
    default_paths.cred_filepath = cred_filepath
    default_params.wiki_domain = wiki_domain
    # This import will trigger the install_dependencies method call on the first run
    import osw.express

    osw_express = osw.express.OswExpress(domain=wiki_domain)
    osw_express_and_credentials(osw_express, wiki_domain, wiki_username, wiki_password)
    assert osw_express.cred_filepath == Path(cred_filepath)
    osw_express.shut_down()
    cred_filepath.unlink()


def test_init_with_cred_filepath(wiki_domain, wiki_username, wiki_password):
    """Test OswExpress initialization with defined domain and cred_filepath."""
    cred_filepath = Path.cwd() / "accounts.pwd.yaml"
    create_credentials_file(cred_filepath, wiki_domain, wiki_username, wiki_password)
    import osw.express

    osw_express = osw.express.OswExpress(
        domain=wiki_domain, cred_filepath=cred_filepath
    )
    assert osw_express_and_credentials(
        osw_express, wiki_domain, wiki_username, wiki_password
    )
    assert osw_express.cred_filepath == cred_filepath
    osw_express.shut_down()
    cred_filepath.unlink()


def test_init_with_cred_filepath_but_missing_credentials(
    wiki_domain, wiki_username, wiki_password, mocker
):
    """Test OswExpress initialization with defined domain and cred_filepath, but missing
    credentials."""
    mocked_input = mocker.patch("builtins.input")
    mocked_getpass = mocker.patch("getpass.getpass")
    mocked_input.return_value = wiki_username
    mocked_getpass.return_value = wiki_password
    cred_filepath = Path.cwd() / "osw_files" / "accounts.pwd.yaml"
    if not cred_filepath.parent.exists():
        cred_filepath.parent.mkdir(parents=True)
    with open(cred_filepath, "w") as f:
        yaml.dump({"dummy.domain": {"username": "dummy", "password": "password"}}, f)
    import osw.express

    osw_express = osw.express.OswExpress(
        domain=wiki_domain, cred_filepath=cred_filepath
    )
    assert osw_express_and_credentials(
        osw_express, wiki_domain, wiki_username, wiki_password
    )
    assert osw_express.cred_filepath == cred_filepath
    osw_express.shut_down()
    cred_filepath.unlink()


def test_init_with_cred_mngr(wiki_domain, wiki_username, wiki_password):
    """Test OswExpress initialization with defined domain and cred_mngr."""
    from osw.auth import CredentialManager

    cred_filepath = Path.cwd() / "accounts.pwd.yaml"
    create_credentials_file(cred_filepath, wiki_domain, wiki_username, wiki_password)
    cred_mngr = CredentialManager(cred_filepath=cred_filepath)
    import osw.express

    osw_express = osw.express.OswExpress(domain=wiki_domain, cred_mngr=cred_mngr)
    assert osw_express_and_credentials(
        osw_express, wiki_domain, wiki_username, wiki_password
    )
    assert osw_express.cred_filepath == cred_filepath
    osw_express.shut_down()
    cred_filepath.unlink()


def test_file_upload_download(wiki_domain, wiki_username, wiki_password):
    cred_filepath = Path.cwd() / "accounts.pwd.yaml"
    create_credentials_file(cred_filepath, wiki_domain, wiki_username, wiki_password)
    import osw.express

    osw_express = osw.express.OswExpress(
        domain=wiki_domain, cred_filepath=cred_filepath
    )
    assert osw_express_and_credentials(
        osw_express, wiki_domain, wiki_username, wiki_password
    )
    # Test file upload
    file_path = Path.cwd() / "test_file.txt"
    create_dummy_file(file_path)
    # Upload a file to an OSW instance
    wiki_file = osw.express.osw_upload_file(file_path, osw_express=osw_express)
    local_file = osw.express.osw_download_file(wiki_file.url, osw_express=osw_express)
    assert local_file.path.exists()
    assert local_file.path.is_file()
    assert local_file.path.read_text() == "Hello, World!"
    assert local_file.url_or_title == wiki_file.url_or_title
    assert local_file.name == wiki_file.name
    # assert local_file.label[0].text == wiki_file.label[0].text  # expected behavior?
    local_file.close()
    local_file.delete()  # Delete downloaded file from local filesystem
    assert local_file.path.exists() is False
    wiki_file.delete()  # Delete WikiFile from OSW instance
    wiki_file.path.unlink()  # Delete original file from local filesystem
    assert wiki_file.path.exists() is False
    # todo: additional test required for setting / overwriting properties of the
    #  WikiFile


if __name__ == "__main__":
    cred_filepath_ = Path("accounts.pwd.yaml")
    with open(cred_filepath_, "r") as f:
        accounts = yaml.safe_load(f)
    wiki_domain_ = "wiki-dev.open-semantic-lab.org"
    wiki_username_ = accounts[wiki_domain_]["username"]
    wiki_password_ = accounts[wiki_domain_]["password"]
    use_pytest = False
    if use_pytest:
        # Test that require the pytest-mock fixture:
        import pytest

        # Runs all tests
        pytest.main(
            [
                __file__,
                "--wiki_domain",
                wiki_domain_,
                "--wiki_username",
                wiki_username_,
                "--wiki_password",
                wiki_password_,
            ]
        )
    else:
        # Runs tests that can be run directly:
        test_init_with_cred_filepath(wiki_domain_, wiki_username_, wiki_password_)
        test_init_with_cred_mngr(wiki_domain_, wiki_username_, wiki_password_)
        test_file_upload_download(wiki_domain_, wiki_username_, wiki_password_)



# This file is required to make Python treat directories containing the file as
#  packages.
# Additionally, classes from multiple files are imported to allow easy access via
#  'osw.controller.<ClassName>

import inspect

import osw.model.entity as model

# imports of controllers for always present models
# note: to keep this import auto-removal per pre-commit hook is disabled
from osw.controller.entity import Entity  # noqa: F401

# conditional imports
# Note: does not work if the module is imported before the model is loaded
models = [m[0] for m in inspect.getmembers(model)]

if "Database" in models:
    from osw.controller.database import DatabaseController  # noqa: F401



import os
import tempfile

import yaml

from osw.auth import CredentialManager


def test_cred_mngr():
    data = {"test.domain.com": {"username": "testuser", "password": "pass123"}}
    data2 = {
        "test.domain.com:80": {"username": "testuser2", "password": "pass1234"},
        "domain.com:80": {"username": "testuser23", "password": "pass12345"},
        "other.domain.com:80": {"username": "h23123", "password": "p3454"},
    }
    with tempfile.TemporaryDirectory() as td:
        f_name = os.path.join(td, "test.yaml")
        with open(f_name, "w") as credfile:
            yaml.dump(data, credfile, default_flow_style=False)
        f_name2 = os.path.join(td, "test2.yaml")
        with open(f_name, "w") as credfile:
            yaml.dump(data, credfile, default_flow_style=False)
        with open(f_name2, "w") as credfile2:
            yaml.dump(data2, credfile2, default_flow_style=False)

        cm = CredentialManager(cred_filepath=credfile.name)
        cred = cm.get_credential(CredentialManager.CredentialConfig(iri="domain.com"))
        assert (
            cred.username == data["test.domain.com"]["username"]
            and cred.password == data["test.domain.com"]["password"]
        )

        cm = CredentialManager(cred_filepath=[credfile.name, credfile2.name])
        cred = cm.get_credential(
            CredentialManager.CredentialConfig(iri="domain.com:80")
        )
        assert (
            cred.username == data2["domain.com:80"]["username"]
            and cred.password == data2["domain.com:80"]["password"]
        )

        extra_cred = CredentialManager.UserPwdCredential(
            username="testuserX", password="pass123X", iri="domainX.com"
        )
        cm.add_credential(extra_cred)
        cred = cm.get_credential(CredentialManager.CredentialConfig(iri="domainX.com"))
        assert (
            cred.username == extra_cred.username
            and cred.password == extra_cred.password
        )



from osw.auth import CredentialManager  # noqa: F401
from osw.core import (  # noqa: F401
    OSW,
    OVERWRITE_CLASS_OPTIONS,
    AddOverwriteClassOptions,
    OverwriteOptions,
)
from osw.express import DataModel, OswExpress  # noqa: F401
from osw.wiki_tools import SearchParam  # noqa: F401
from osw.wtsite import WtPage, WtSite  # noqa: F401


# From osw.auth.CredentialManager:
class UserPwdCredential(CredentialManager.UserPwdCredential):
    pass


# Enum
# class CredentialFallback(CredentialManager.CredentialFallback):
#     pass


class CredentialConfig(CredentialManager.CredentialConfig):
    pass


# From osw.core.OSW:
class SchemaUnregistration(OSW.SchemaUnregistration):
    pass


class SchemaRegistration(OSW.SchemaRegistration):
    pass


# Enum
# class FetchSchemaMode(OSW.FetchSchemaMode):
#     pass


class FetchSchemaParam(OSW.FetchSchemaParam):
    pass


class LoadEntityParam(OSW.LoadEntityParam):
    pass


class OverwriteClassParam(OSW.OverwriteClassParam):
    pass


class StoreEntityParam(OSW.StoreEntityParam):
    pass


class DeleteEntityParam(OSW.DeleteEntityParam):
    pass


class QueryInstancesParam(OSW.QueryInstancesParam):
    pass


# From osw.wtsite.WtSite:
class WtSiteConfig(WtSite.WtSiteConfig):
    pass


class GetPageParam(WtSite.GetPageParam):
    pass


class ModifySearchResultsParam(WtSite.ModifySearchResultsParam):
    pass


class UploadPageParam(WtSite.UploadPageParam):
    pass


class CopyPagesParam(WtSite.CopyPagesParam):
    pass


class CreatePagePackageParam(WtSite.CreatePagePackageParam):
    pass


class ReadPagePackageParam(WtSite.ReadPagePackageParam):
    pass


class UploadPagePackageParam(WtSite.UploadPagePackageParam):
    pass


class DeletePageParam(WtSite.DeletePageParam):
    pass


# From osw.wtsite.WtPage:
class CopyPageConfig(WtPage.CopyPageConfig):
    pass


class PageDumpConfig(WtPage.PageDumpConfig):
    pass


class ExportConfig(WtPage.ExportConfig):
    pass


class ImportConfig(WtPage.ImportConfig):
    pass



import os

import osw.model.entity as model
from osw.express import OswExpress

# Create/update the password file under examples/accounts.pwd.yaml
pwd_file_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
)
# pwd_file_path = "./accounts.pwd.yaml"
osw_obj = OswExpress(
    domain="wiki-dev.open-semantic-lab.org", cred_filepath=pwd_file_path
)

# Load Tutorial Schema on demand
if not hasattr(model, "Tutorial"):
    DEPENDENCIES = {"Tutorial": "Category:OSW494f660e6a714a1a9681c517bbb975da"}
    osw_obj.install_dependencies(DEPENDENCIES, mode="replace")

# Load instance HowToCreateANewArticle
title = "Item:OSW52c2c5a6bbc84fcb8eab0fa69857e7dc"
entity = osw_obj.load_entity(title)
print(entity.__class__)
print(entity.label[0].text)  # We can access any attribute of model.Entity

tutorial_entity = entity.cast(model.Tutorial)  # Explicit cast to model.Tutorial
print(
    tutorial_entity.required_predecessor
)  # We can access now any attribute of model.Tutorial

print(entity.json(exclude_none=True))  # export as json

# Load all instances of Tutorial in parallel
tutorials = osw_obj.query_instances(
    category="Category:OSW494f660e6a714a1a9681c517bbb975da"
)
print(tutorials)
osw_obj.load_entity(tutorials)



# generated by datamodel-codegen:
#   filename:  Item.json
#   timestamp: 2024-09-12T12:35:11+00:00

from __future__ import annotations

from typing import Any, List, Literal, Optional, Set, Union
from uuid import UUID, uuid4

from pydantic.v1 import Field, constr

from osw.model.static import OswBaseModel


class ReadAccess(OswBaseModel):
    level: Optional[Literal["public", "internal", "restricted"]] = Field(
        None, title="Level"
    )


class AccessRestrictions(OswBaseModel):
    read: Optional[ReadAccess] = Field(None, title="Read access")


class Label(OswBaseModel):
    text: constr(min_length=1) = Field(..., title="Text")
    lang: Optional[Literal["en", "de"]] = Field("en", title="Lang code")


class Description(Label):
    pass


class WikiPage(OswBaseModel):
    """
    The wiki page containing this entity
    """

    title: Optional[str] = Field(None, title="Title")
    """
    The page title
    """
    namespace: Optional[str] = Field(None, example="Category", title="Namespace")
    """
    The page namespace
    """


class Meta(OswBaseModel):
    uuid: UUID = Field(default_factory=uuid4, title="UUID")
    wiki_page: Optional[WikiPage] = Field(None, title="Wiki page")
    """
    The wiki page containing this entity
    """
    change_id: Optional[List[str]] = Field(None, title="Change IDs")
    """
    To keep track of concerted changes
    """


class Entity(OswBaseModel):
    rdf_type: Optional[Set[str]] = Field(None, title="Additional RDF type(s)")
    """
    Declares additional type(s) for this entity, e.g., to state that this entity has the same meaning as a term in a controlled vocabulary or ontology. This property is synonymous to the schema:additionalType and owl:sameAs. The default syntax is ontology:TermName. The ontology prefix has to be defined in the @context of the Entity, the category or any of the parent categories. The term name has to be a valid identifier in the ontology.
    """
    uuid: UUID = Field(default_factory=uuid4, title="UUID")
    iri: Optional[str] = Field(None, title="IRI")
    """
    The Internationalized Resource Identifier (IRI) of this entity
    """
    name: Optional[str] = Field(None, title="Technical name")
    """
    Technical / Machine compatible name
    """
    label: List[Label] = Field(..., min_items=1, title="Label(s)")
    """
    At least one label is required.
    """
    short_name: Optional[List[Label]] = Field(None, title="Short name(s)")
    """
    Abbreviation, Acronym, etc.
    """
    query_label: Optional[str] = Field(None, title="Query label")
    description: Optional[List[Description]] = Field(None, title="Description")
    image: Optional[str] = Field(None, title="Image")
    ordering_categories: Optional[List[str]] = Field(None, title="Ordering categories")
    """
    Ordering categories are used to categorize instances, e.g., according to their use but not their properties. When querying for instances of a here listed ordering category, this instance will be returned. Note: Ordering categories define no properties, while 'regular' categories define properties, which an instance assigns values to.
    """
    keywords: Optional[List[str]] = Field(None, title="Keywords / Tags")
    """
    Designated to the user defined categorization of this element
    """
    based_on: Optional[List[str]] = Field(None, title="Based on")
    """
    Other entities on which this one is based, e.g. when it is created by copying
    """
    statements: Optional[
        List[Union[ObjectStatement, DataStatement, QuantityStatement]]
    ] = Field(None, title="Statements")
    attachments: Optional[List[str]] = Field(None, title="File attachments")
    meta: Optional[Meta] = None


class ObjectStatement(OswBaseModel):
    rdf_type: Optional[Any] = "rdf:Statement"
    uuid: UUID = Field(default_factory=uuid4, title="UUID")
    label: Optional[List[Label]] = Field(None, title="Label")
    """
    Human readable name
    """
    subject: Optional[str] = Field(None, title="Subject")
    substatements: Optional[
        List[Union[ObjectStatement, DataStatement, QuantityStatement]]
    ] = Field(None, title="Substatements")
    predicate: str = Field(..., title="Predicate")
    object: str = Field(..., title="Object")


class DataStatement(OswBaseModel):
    rdf_type: Optional[Any] = "rdf:Statement"
    uuid: UUID = Field(default_factory=uuid4, title="UUID")
    label: Optional[List[Label]] = Field(None, title="Label")
    """
    Human readable name
    """
    subject: Optional[str] = Field(None, title="Subject")
    substatements: Optional[
        List[Union[ObjectStatement, DataStatement, QuantityStatement]]
    ] = Field(None, title="Substatements")
    property: str = Field(..., title="Property")
    value: str = Field(..., title="Value")


class QuantityStatement(OswBaseModel):
    rdf_type: Optional[Any] = "rdf:Statement"
    uuid: UUID = Field(default_factory=uuid4, title="UUID")
    label: Optional[List[Label]] = Field(None, title="Label")
    """
    Human readable name
    """
    subject: Optional[str] = Field(None, title="Subject")
    substatements: Optional[
        List[Union[ObjectStatement, DataStatement, QuantityStatement]]
    ] = Field(None, title="Substatements")
    quantity: str = Field(..., title="Property")
    numerical_value: str = Field(..., title="Value")
    unit: str = Field(..., title="Unit")
    unit_symbol: str
    value: str = Field(..., title="Value")


class Item(Entity):
    type: Optional[List[str]] = Field(
        ["Category:Item"], min_items=1, title="Types/Categories"
    )
    entry_access: Optional[AccessRestrictions] = Field(
        None, title="Access restrictions"
    )


Entity.update_forward_refs()
ObjectStatement.update_forward_refs()
DataStatement.update_forward_refs()
QuantityStatement.update_forward_refs()
Item.update_forward_refs()



import sys
from pathlib import Path

from pydantic.v1.types import FilePath as PydanticFilePath
from pydantic.v1.validators import path_validator  # , path_exists_validator

if sys.version_info < (3, 10):
    NoneType = type(None)
else:
    from types import NoneType  # noqa: F401


class PossibleFilePath(PydanticFilePath):
    # Overwrite the Pydantic FilePath class to allow non-existing paths
    @classmethod
    def __get_validators__(cls):
        yield path_validator
        # yield path_exists_validator  # Remove this line to allow non-existing paths
        yield cls.validate

    @classmethod
    def validate(cls, value: Path) -> Path:
        return value



from pathlib import Path

import osw.params as prm
from osw.express import OswExpress

osw_obj = OswExpress(
    domain="demo.open-semantic-lab.org", cred_filepath=Path("accounts.pwd.yaml")
)
orgs = osw_obj.site.semantic_search(
    prm.SearchParam(
        query="[[HasType::Category:OSW1969007d5acf40539642877659a02c23]]", debug=False
    )
)
persons = osw_obj.site.semantic_search(
    prm.SearchParam(
        query="[[HasType::Category:OSW44deaa5b806d41a2a88594f562b110e9]]", debug=False
    )
)
osw_obj.site.delete_page(orgs + persons)



import os
from io import StringIO

import osw.model.entity as model
from osw.auth import CredentialManager
from osw.express import OswExpress
from osw.utils.wiki import get_full_title

# Install: pip install osw[S3]

# Credential manager (can use a file or hardcode the credentials, otherwise the user
#  will be prompted to enter them)
cm = CredentialManager(
    cred_filepath=os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
    )
)
# Change the domain to your osw instance
osw_obj = OswExpress(domain="wiki-dev.open-semantic-lab.org", cred_mngr=cm)

# Load the required schemas / data classes
if not hasattr(model, "S3File"):
    DEPENDENCIES = {
        "S3File": "Category:OSWc43f749badcb4490a785505de1fc7d20",
        "LocalFile": "Category:OSW3e3f5dd4f71842fbb8f270e511af8031",
        "WikiFile": "Category:OSW11a53cdfbdc24524bf8ac435cbf65d9d",
    }
    osw_obj.install_dependencies(DEPENDENCIES, mode="replace")

# Import the controller modules
# Note: since they depend on the data classes, they must be imported after the schemas
#  are loaded
from osw.controller.file.local import (  # noqa (ignore flake8 warning)
    LocalFileController,
)
from osw.controller.file.s3 import S3FileController  # noqa (ignore flake8 warning)
from osw.controller.file.wiki import WikiFileController  # noqa (ignore flake8 warning)


def video_file():
    file = osw_obj.load_entity(
        "File:OSW7b2398a60f004006b9a6ef89210858f3.mp4"
    )  # the file
    wf2 = file.cast(WikiFileController, osw=osw_obj)  # the file controller
    # wf2.suffix = "mp4"
    lf2 = LocalFileController.from_other(wf2, path="dummy2.mp4")
    lf2.put_from(wf2)


def wiki_file():
    # Create a local file
    # with open("dummy.txt", "w") as f:
    #    f.write("Hello World!")
    lf = LocalFileController(path="dummy.txt")  # here an uuid already exists
    lf.put(StringIO("Hello World!"))

    # Create a remote file (here: a wiki file)
    wf = WikiFileController(osw=osw_obj)
    # Or cast to wiki wile to keep all common attributes
    wf = lf.cast(WikiFileController, osw=osw_obj)
    # Which is equivalent to
    wf = WikiFileController.from_other(lf, osw=osw_obj)

    # Upload the local file to the remote file
    wf.put_from(lf)
    # Write new content on the fly
    wf.put(StringIO("Some new content"))

    # Get an existing file
    file = osw_obj.load_entity(f"{wf.namespace}:{wf.title}")  # the file
    wf2 = file.cast(WikiFileController, osw=osw_obj)  # the file controller
    lf2 = LocalFileController.from_other(wf2, path="dummy2.txt")
    lf2.put_from(wf2)
    get_full_title(wf2)

    # Delete the files
    lf.delete()
    wf.delete()  # note: wf2 actually points to the same file as wf
    lf2.delete()


def s3_file():
    # Make sure your credential file contains a access_key_id=username,
    #  and secret_access_key=password for the s3 domain
    file = osw_obj.load_entity("Item:OSW5f53ed0b5c354fc3b1a122b9066744f3")  # the file
    s3f = file.cast(S3FileController, cm=cm)  # the file controller

    # Download
    lf = LocalFileController.from_other(s3f, path="s3_test.txt")
    lf.put_from(s3f)

    # Upload
    s3f.put_from(lf)

    s3f = model.S3File(
        url="https://s3.example10.open-semantic-lab.org/test-bucket/test-example2.txt",
        label=[model.Label(text="NewFile")],
    ).cast(S3FileController, cm=cm)
    s3f.put(StringIO("Hello World!"))

    s3f.delete()



import importlib.util
import json
import re
import sys
from pathlib import Path
from typing import Optional, Union
from warnings import warn

from pydantic.v1 import FilePath
from typing_extensions import Dict, List

import osw.model.page_package as model
from osw.auth import CredentialManager
from osw.model import page_package as package
from osw.model.page_package import NAMESPACE_CONST_TO_NAMESPACE_MAPPING
from osw.model.static import OswBaseModel
from osw.utils.regex import RegExPatternExtended
from osw.wtsite import WtSite

# Definition of constants
PATTERNS = {
    "Category string with prefix in quotes": RegExPatternExtended(
        description='Match: "Category:<>"',
        pattern=r".*\"(Category:.+)\"",
        group_keys=["category"],
    ),
    "Category string with prefix in context and allOf": RegExPatternExtended(
        description="Match: /wiki/Category:<>?action",
        pattern=r"./wiki\/(Category:.+)\?action",
        group_keys=["category"],
    ),
    "Category string with prefix in brackets": RegExPatternExtended(
        description="Match: [[Category:<>]] and [[HasType::Category:<>]]",
        pattern=r"\[\[.*(Category:[a-zA-z0-9]+)\]\]",
        group_keys=["category"],
    ),
    "Property sting with prefix in quotes": RegExPatternExtended(
        description='Match: "Property:<>"',
        pattern=r"\"(Property:.+)\"",
        group_keys=["property"],
    ),
    "Property string with prefix and @id": RegExPatternExtended(
        description='Match: "@id": "Property:<>"',
        pattern=r'"@id":\s*\"(Property:.+)\"',
        group_keys=["property"],
    ),
    "Listed properties": RegExPatternExtended(
        description='Match: properties: ["<>", "<>"] and '
        'ignore_properties: ["<>", "<>"]',
        pattern=r"\"*properties\":\s*\[(\"[^\].]*\")\]",
        group_keys=["property"],
    ),
    "Property strings in autocomplete queries": RegExPatternExtended(
        description="Match: [[<Property>::Category:<Category>]]",
        pattern=r"\[\[([a-zA-Z\_]+)::.*\]\]",
        group_keys=["property"],
    ),
    "Property strings in general queries": RegExPatternExtended(
        description="Match: |?<Property>= and |<Property>=",
        pattern=r"\|[\?]*\s*([^\?^\=^#^\n^\{^\}^>^<^@]+)[\s\=\|\}]+",
        group_keys=["property"],
    ),
    "Template strings": RegExPatternExtended(
        description="Match: /wiki/JsonSchema:Label=action",
        pattern=r"\"\/wiki\/(.*:.*)\?action",
        group_keys=["template"],
    ),
    "File strings": RegExPatternExtended(
        description="Match: File:OSW<uuid>.<suffix>",
        pattern=r"(File:OSW[a-f0-9]{32}\.[^\\^\/^\"^\s]+)",
        group_keys=["file"],
    ),
    "Item strings": RegExPatternExtended(
        description="Match: Item:OSW<uuid>",
        pattern=r"(Item:OSW[a-f0-9]{32})",
        group_keys=["item"],
    ),
}

PATTERNS_PER_SLOT = {
    "jsondata": [
        PATTERNS["Category string with prefix in quotes"],
        PATTERNS["Property sting with prefix in quotes"],
        PATTERNS["Item strings"],
        PATTERNS["File strings"],
    ],
    "jsonschema": [
        PATTERNS["Category string with prefix in context and allOf"],
        PATTERNS["Category string with prefix in quotes"],
        PATTERNS["Property strings in autocomplete queries"],
        PATTERNS["Property strings in general queries"],
        PATTERNS["Property sting with prefix in quotes"],
        PATTERNS["Template strings"],
        PATTERNS["Item strings"],
        PATTERNS["File strings"],
    ],
}
"""
    "footer_template": [
        PATTERNS["Listed properties"],
        PATTERNS["Property strings in autocomplete queries"],
        # This may cause way to many info box labels to end up in
        # the list of requirements:
        # PATTERNS["Property strings in general queries"],

    ],
    "header_template": [
        PATTERNS["Listed properties"],
        PATTERNS["Property strings in autocomplete queries"],
        # This may cause way to many info box labels to end up in
        # the list of requirements:
        # PATTERNS["Property strings in general queries"],
    ],
    "main": [
        PATTERNS["Category string with prefix in brackets"],
    ]
}
"""

EXCEPTIONS = [
    "de",
    "en",
    "#default",
    "format",
    "machine compatible name",
    "label",
    "short name",
    "mainlabel",
    "template",
    "named args",
    "Property",
    ",",
]

EXCEPTION_PATTERNS = {
    "Contains no capital letter": RegExPatternExtended(
        description="Match e.g.: 200x800px]]|",
        pattern=r"^[^A-Z]*$",
        group_keys=[],  # no group keys
    ),
}
"""
    "Contains backward slashes": RegExPatternExtended(
        description='Match: \\',
        pattern=r'\\',
        group_keys=[],  # no group keys
    ),
    "Contains forward slashes": RegExPatternExtended(
        description='Match: /',
        pattern='\\/',
        group_keys=[],  # no group keys
    ),
    "Contains brackets": RegExPatternExtended(
        description='Match e.g.: [[Category:OSWfe72974590fd4e8ba94cd4e8366375e8]]',
        pattern=r'[\\[\\]\\(\\)\\{\\}]+',
        group_keys=[],  # no group keys
    ),
    "Contains pipe char": RegExPatternExtended(
        description='Match e.g.: |',
        pattern=r'\\|',
        group_keys=[],  # no group keys
    ),
    "Contains quotation marks": RegExPatternExtended(
        description='Match e.g.: "',
        pattern=r'\\"',
        group_keys=[],  # no group keys
    ),
    "Contains hash": RegExPatternExtended(
        description='Match e.g.: "',
        pattern=r'#',
        group_keys=[],  # no group keys
    ),
}
"""


# Definition of functions
def import_module_from_file(module_name: str, file_path: Path):
    """Imports a module from a file. THis is usefull when the module
    has "." in the name!"""
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


def read_package_script_file(script_fp: Path, package_name: str = None) -> dict:
    """Reads the package info from a script file"""
    if package_name is None:
        package_name = script_fp.stem.replace(".", "_").replace("-", "_")
    package_module = import_module_from_file(package_name, script_fp)
    return {
        "package_meta_data": package_module.package_meta_data,
        "package_creation_config": package_module.package_creation_config,
        "package_creation_script_fp": script_fp,
    }


def get_listed_pages_from_package_script(package_script: dict) -> List[str]:
    """Takes in the output of read_package_script_file and returns a list of
    pages listed in the package"""
    return getattr(package_script["package_meta_data"], "page_titles", [])


def get_required_packages_from_package_script(package_script: dict) -> List[str]:
    """Takes in the output of read_package_script_file and returns a list of
    packages listed in the package"""
    return getattr(package_script["package_meta_data"], "requiredPackages", [])


def read_package_info_file(
    package_dir: Path, package_info_fn: str = "packages.json"
) -> dict:
    """Reads the package info from a file in the package directory"""
    matching_files_in_dir = list(package_dir.glob(package_info_fn))
    if len(matching_files_in_dir) == 0:
        raise FileNotFoundError(f"No file {package_info_fn} found in {package_dir}")
    if len(matching_files_in_dir) > 1:
        raise ValueError(
            f"Multiple files {package_info_fn} found in {package_dir}: {matching_files_in_dir}"
        )
    else:
        package_info_fp = matching_files_in_dir[0]
        with open(package_info_fp, "r") as package_info_fp:
            package_info = json.load(package_info_fp)
        return package_info


def find_package_dir(
    package_or_script_name: str, search_paths: List[Path] = None
) -> Path:
    """searches for the subfolder or script name with
    name <package_or_script_name> in the search_paths"""
    if search_paths is None:
        search_paths = []
    matching_dirs = []
    for search_path in search_paths:
        matching_dirs.extend(list(search_path.glob(f"{package_or_script_name}")))
    if len(matching_dirs) == 0:
        raise FileNotFoundError(f"{package_or_script_name} not found in {search_paths}")
    if len(matching_dirs) > 1:
        raise ValueError(
            f"""Multiple elements {package_or_script_name} found in
            {search_paths}: {matching_dirs}"""
        )
    else:
        return matching_dirs[0]


def get_listed_pages_from_package_info(package_info: Union[dict, Path]) -> List[str]:
    """Takes in the output of read_package_info_file and returns a list of
    pages listed in the package"""
    if isinstance(package_info, Path):
        package_info = read_package_info_file(package_info)
    listed_pages = []
    for val in package_info["packages"].values():
        for page in val["pages"]:
            if page.get("namespace"):
                namespace = NAMESPACE_CONST_TO_NAMESPACE_MAPPING.get(page["namespace"])
            else:
                namespace = ""
            listed_pages.append(f"{namespace}:{page['name']}")
    return listed_pages


def get_required_packages_from_package_info_file(package_info: dict) -> List[str]:
    """Takes in the output of read_package_info_file and returns a list of
    packages listed in the package"""
    required_packages = []
    for val in package_info["packages"].values():
        if val.get("requiredPackages"):
            required_packages.extend(val["requiredPackages"])
    return required_packages


def check_for_exceptions(string: str) -> bool:
    """Checks if a string is listed as an exception or falls into
    an exception pattern."""
    is_exception = False
    # Check for empty string
    if len(string) == 0:
        return True
    for pattern in EXCEPTION_PATTERNS.values():
        if re.search(pattern.pattern.pattern, string):
            return True
    for exception_string in EXCEPTIONS:
        if string.lower() == exception_string:
            return True
    return is_exception


def get_required_pages_from_file(fp: Union[str, Path]) -> List[str]:
    """Returns a list of pages required by the package"""

    required_pages = []
    # suffix = fp.suffix
    try:
        pre_suffix = fp.suffixes[-2]
    except IndexError:
        pre_suffix = ".slot_main"

    slot_name = pre_suffix.split(".slot_")[-1]
    with open(fp, "r", encoding="utf-8") as f:
        content = f.read()
    if not PATTERNS_PER_SLOT.get(slot_name):
        return required_pages
    for pattern in PATTERNS_PER_SLOT.get(slot_name):
        match_res = re.findall(pattern.pattern.pattern, content, re.MULTILINE)
        if match_res:
            if "category" in pattern.group_keys:
                for match in match_res:
                    required_pages.append(match)
            elif "property" in pattern.group_keys:
                for match in match_res:
                    if '"' in match:
                        results = match.split('"')
                    else:
                        results = [match]
                    for res in results:
                        intermediate = []
                        if "-" in res:
                            intermediate.append(res.split("-")[-1])
                        elif "." or ":" in res:
                            intermediate.extend(res.split("."))
                            # intermediate.extend(res.split(":"))
                        else:
                            intermediate.append(res)
                        required_pages.extend(
                            [
                                "Property:" + s
                                for s in intermediate
                                if not s.startswith("Property:")
                                and not check_for_exceptions(s)
                                # only here to catch exceptions early
                            ]
                        )
            elif "template" in pattern.group_keys:
                for match in match_res:
                    required_pages.append(match)
            elif "item" in pattern.group_keys:
                for match in match_res:
                    required_pages.append(match)
            elif "file" in pattern.group_keys:
                for match in match_res:
                    required_pages.append(match)
    return_obj = list(
        set(entry for entry in required_pages if not check_for_exceptions(entry))
    )
    return_obj.sort()
    return return_obj


# Definition of classes
class PagePackageController(model.PagePackageMetaData):
    """Adds controller logic to a package definitions to create / update it"""

    class CreationConfig(OswBaseModel):
        """Configuration for creating a page package. This is the data needed to create
        a page package but is not included in the page package."""

        domain: str
        """A string formatted as domain"""
        cred_filepath: Union[str, FilePath]
        """Path to a existing credentials yaml files"""
        working_dir: Union[str, Path]
        """Working directory. Will be created automatically if not existing."""
        skip_slot_suffix_for_main: bool = False
        ignore_titles: Optional[List[str]] = None
        """List of page titles to ignore in auto-detected dependencies, e.g.,
        example files"""

    def create(
        self,
        creation_config: CreationConfig,
    ):
        """creates or updates the package

        Parameters
        ----------
        creation_config
            see PagePackageController.CreationConfig
        """
        # Create a WtSite instance to load pages from the specified domain
        wtsite = WtSite(
            WtSite.WtSiteConfig(
                iri=creation_config.domain,
                cred_mngr=CredentialManager(
                    cred_filepath=creation_config.cred_filepath
                ),
            )
        )
        # Create a PagePackageBundle instance
        bundle = package.PagePackageBundle(
            publisher=self.publisher,
            author=self.author,
            language=self.language,
            publisherURL=f"https://github.com/{self.repo_org}/{self.repo}",
            packages={
                f"{self.name}": package.PagePackage(
                    globalID=f"{self.id}",
                    label=self.name,
                    version=self.version,
                    requiredExtensions=self.requiredExtensions,
                    requiredPackages=self.requiredPackages,
                    description=self.description,
                    baseURL=f"https://raw.githubusercontent.com/"
                    f"{self.repo_org}/"
                    f"{self.repo}/"
                    f"{self.branch}/"
                    f"{self.subdir}/",
                )
            },
        )
        # Create a PagePackageConfig instance
        config = package.PagePackageConfig(
            name=self.name,
            config_path=Path(creation_config.working_dir) / "packages.json",
            content_path=Path(creation_config.working_dir) / self.subdir,
            bundle=bundle,
            titles=self.page_titles,
            ignore_titles=creation_config.ignore_titles,
            skip_slot_suffix_for_main=creation_config.skip_slot_suffix_for_main,
        )
        # Create the page package in the working directory
        wtsite.create_page_package(WtSite.CreatePagePackageParam(config=config))

    class GetRequiredPagesParams(OswBaseModel):
        creation_config: "PagePackageController.CreationConfig"
        direct_call: bool = True
        """Whether this function is called directly"""

    def get_required_pages(
        self, params: GetRequiredPagesParams
    ) -> Dict[str, Union[List[str], Dict[str, List[str]]]]:
        """Returns a list of pages required by the package"""

        required_pages_path = (
            Path(params.creation_config.working_dir) / "required_pages.json"
        )
        required_pages_dict = {}
        if required_pages_path.exists() and not params.direct_call:
            with open(required_pages_path, "r") as f:
                self.requiredPages = json.load(f)["list"]
        else:
            self.requiredPages = []
            content_path = Path(params.creation_config.working_dir) / self.subdir
            list_of_files = list(content_path.glob("**/*"))
            # parse the files
            for title in self.page_titles:
                required_pages_dict[title] = {}
                title_specific_files = [
                    file for file in list_of_files if title.split(":")[-1] in file.name
                ]
                for file in title_specific_files:
                    # For now: skip non (!) wikitext and json files
                    if file.suffix not in [".wikitext", ".json"]:
                        continue
                    try:
                        pre_suffix = file.suffixes[-2]
                    except IndexError:
                        pre_suffix = ".slot_main"
                    slot_name = pre_suffix.split(".slot_")[-1]
                    required_pages = get_required_pages_from_file(file)
                    """File specific required pages"""
                    self.requiredPages.extend(required_pages)
                    if required_pages:
                        required_pages_dict[title][slot_name] = required_pages
                if not required_pages_dict[title]:  # test if empty
                    del required_pages_dict[title]
            # Remove duplicates
            self.requiredPages = list(set(self.requiredPages))
        self.requiredPages.sort()
        return_obj = {"list": self.requiredPages, "dict": required_pages_dict}
        with open(required_pages_path, "w", encoding="utf-8") as f:
            json.dump(return_obj, f, indent=4, ensure_ascii=False)

        return return_obj

    class CheckRequiredPagesParams(OswBaseModel):
        creation_config: "PagePackageController.CreationConfig"
        direct_call: bool = True
        """Whether this function is called directly and therefore to call
        get_required_pages() before checking for missing pages"""
        read_listed_pages_from_script: bool = False
        """Whether to read the listed pages and package info from the script
        file or from the package info file (default: packages.json)"""
        script_dir: Union[str, Path] = None
        """Path to the directory where the package creation script is stored.
        Only required if read_listed_pages_from_script is True"""
        label_missing_pages: bool = None
        """Whether to get the labels of the missing pages from the OSW"""
        additional_script_dirs: List[Union[str, Path]] = None
        """Additional directories to search for management scripts of dependencies
        Note: This may lead to errors for relative imports"""
        additional_package_dirs: List[Union[str, Path]] = None
        """Additional directories to search for dependencies"""

        # class Config:
        #     arbitrary_types_allowed = True

    def check_required_pages(
        self,
        params: CheckRequiredPagesParams,
    ):
        """Checks if all required pages are listed in the package and packages
        listed as requiredPackages. Simultaneously checks for redundant pages
        along the dependency chain.
        """
        if params.label_missing_pages is None:
            params.label_missing_pages = params.direct_call

        if params.additional_package_dirs is None:
            params.additional_package_dirs = []
        if params.additional_script_dirs is None:
            params.additional_script_dirs = []

        # Set self.requiredPages (either load from file or generate list)
        if params.direct_call:
            self.get_required_pages(
                PagePackageController.GetRequiredPagesParams(**params.dict())
            )
        # For debugging the recursive function
        rec_count = 0

        def recursive(
            package_to_process: str,
            processed_packages: List[str],
            listed_pages: Dict[str, List[str]],
            redundant_pages: Dict[str, List[str]],
        ):
            nonlocal rec_count
            rec_count += 1
            # else:
            if params.read_listed_pages_from_script:
                search_paths = [params.script_dir]
                search_paths.extend(params.additional_script_dirs)
                try:
                    script_path = find_package_dir(
                        f"{package_to_process}.py", search_paths
                    )
                    package_script = read_package_script_file(script_path)
                    new_listed_pages = get_listed_pages_from_package_script(
                        package_script
                    )
                    required_packages = get_required_packages_from_package_script(
                        package_script
                    )
                except Exception as e:
                    warn(f"Error reading package script for {package_to_process}: {e}")
                    new_listed_pages = []
                    required_packages = []
            else:
                search_paths = [params.creation_config.working_dir.parent]
                search_paths.extend(params.additional_package_dirs)
                try:
                    package_dir = find_package_dir(package_to_process, search_paths)
                    package_info = read_package_info_file(package_dir)
                    new_listed_pages = get_listed_pages_from_package_info(package_info)
                    required_packages = get_required_packages_from_package_info_file(
                        package_info
                    )
                except Exception as e:
                    warn(f"Error reading package info for {package_to_process}: {e}")
                    new_listed_pages = []
                    required_packages = []
            # Check for redundant pages
            for pack_ in listed_pages.keys():
                new_redundant_pages = list(
                    set(new_listed_pages) & set(listed_pages[pack_])
                )
                for page_ in new_redundant_pages:
                    if redundant_pages.get(page_):
                        if pack_ not in redundant_pages[page_]:
                            redundant_pages[page_].append(package_to_process)
                    else:
                        redundant_pages[page_] = [pack_, package_to_process]
            # Add new listed pages to listed_pages
            listed_pages[package_to_process] = new_listed_pages
            # Document that this package has been processed
            processed_packages.append(package_to_process)
            # Prepare the return value
            return_val = {
                "processed_packages": processed_packages,
                "listed_pages": listed_pages,
                "redundant_pages": redundant_pages,
            }
            # Recursion Base case
            if len(required_packages) == 0:
                return return_val
            else:  # Recursion
                for pack_ in required_packages:
                    # Don't process packages that have already been processed
                    if pack_ in processed_packages:
                        continue
                    ret = recursive(
                        package_to_process=pack_,
                        processed_packages=processed_packages,
                        listed_pages=listed_pages,
                        redundant_pages=redundant_pages,
                    )
                    # Combine the returns
                    for pack__ in ret["listed_pages"].keys():
                        if return_val["listed_pages"].get(pack__):
                            return_val["listed_pages"][pack__].extend(
                                ret["listed_pages"][pack__]
                            )
                        else:
                            return_val["listed_pages"][pack__] = ret["listed_pages"][
                                pack__
                            ]
                    for pack__ in ret["processed_packages"]:
                        if pack__ not in return_val["processed_packages"]:
                            return_val["processed_packages"].append(pack__)
                    for page__ in ret["redundant_pages"].keys():
                        if return_val["redundant_pages"].get(page__):
                            for pack__ in ret["redundant_pages"][page__]:
                                if pack__ not in return_val["redundant_pages"][page__]:
                                    return_val["redundant_pages"][page__].append(pack__)
                        else:
                            return_val["redundant_pages"][page__] = ret[
                                "redundant_pages"
                            ][page__]

                return return_val

        # Call recursive function
        rec_ret = recursive(
            package_to_process=self.repo,
            processed_packages=[],
            listed_pages={},
            redundant_pages={},
        )
        # print(f"Recursive function called {rec_count} times.")
        # Report on redundantly listed pages
        for pg in rec_ret["redundant_pages"].keys():
            pks = rec_ret["redundant_pages"][pg]
            warn(f"Page {pg} is listed in {len(pks)} packages!: {pks}")
        # Get all listed pages
        all_listed_pages = []
        for pck in rec_ret["listed_pages"].keys():
            all_listed_pages.extend(rec_ret["listed_pages"][pck])
        # Check if all required pages are listed
        missing_pages = list(set(self.requiredPages) - set(all_listed_pages))
        missing_pages.sort()
        # Save the missing pages to a file
        if params.direct_call:
            with open(
                file=Path(params.creation_config.working_dir) / "missing_pages.json",
                mode="w",
                encoding="utf-8",
            ) as f:
                json.dump(missing_pages, f, indent=4, ensure_ascii=False)

        if params.label_missing_pages:
            # Reading missing page labels from file if it already exists
            missing_pages_labeled_fp = (
                Path(params.creation_config.working_dir) / "missing_pages_labeled.json"
            )
            prev_missing_page_labels = {}
            if missing_pages_labeled_fp.exists():
                with open(missing_pages_labeled_fp, "r", encoding="utf-8") as f:
                    prev_missing_page_labels = json.load(f)
            # Create a new dict for the labels of missing pages to avoid unwanted
            # migration of "old" keys
            missing_pages_labeled = {}
            # Only query pages that are not yet labeled
            pages_to_query = list(
                set(missing_pages) - set(list(prev_missing_page_labels.keys()))
            )  # pages that are not yet labeled

            found_pages = {}
            if not len(pages_to_query) == 0:
                # Get the labels of the missing pages from the OSW
                wtsite_obj = WtSite(
                    WtSite.WtSiteConfig(
                        iri=params.creation_config.domain,
                        cred_mngr=CredentialManager(
                            cred_filepath=params.creation_config.cred_filepath
                        ),
                    )
                )
                # Query pages
                pages = wtsite_obj.get_page(
                    WtSite.GetPageParam(titles=pages_to_query)
                ).pages
                # Process returned pages
                for page in pages:
                    title = page.title
                    jsondata = page.get_slot_content("jsondata")
                    if jsondata is None:
                        label = "Page has no jsondata."
                    else:
                        label = "Page has no label."
                        if jsondata.get("label"):
                            label = jsondata["label"][0]["text"]
                    found_pages[title] = label

            # Add found pages to missing_pages_labeled_new
            for page_ in missing_pages:
                if prev_missing_page_labels.get(page_):
                    missing_pages_labeled[page_] = prev_missing_page_labels[page_]
                elif found_pages.get(page_):
                    missing_pages_labeled[page_] = found_pages[page_]
                else:
                    missing_pages_labeled[page_] = (
                        f"Page not found in {params.creation_config.domain}."
                    )
            with open(missing_pages_labeled_fp, "w", encoding="utf-8") as f:
                json.dump(missing_pages_labeled, f, indent=4, ensure_ascii=False)


PagePackageController.CheckRequiredPagesParams.update_forward_refs()
PagePackageController.GetRequiredPagesParams.update_forward_refs()
